# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2017 Maarten de Vries <maarten@de-vri.es>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Autosort automatically keeps your buffers sorted and grouped by server.
# You can define your own sorting rules. See /help autosort for more details.
#
# https://github.com/de-vri-es/weechat-autosort
#

#
# Changelog:
# 3.6:
#   * Add more documentation on provided info hooks.
# 3.5:
#   * Add ${info:autosort_escape,...} to escape arguments for other info hooks.
# 3.4:
#   * Fix rate-limit of sorting to prevent high CPU load and lock-ups.
#   * Fix bug in parsing empty arguments for info hooks.
#   * Add debug_log option to aid with debugging.
#   * Correct a few typos.
# 3.3:
#   * Fix the /autosort debug command for unicode.
#   * Update the default rules to work better with Slack.
# 3.2:
#   * Fix python3 compatiblity.
# 3.1:
#   * Use colors to format the help text.
# 3.0:
#   * Switch to evaluated expressions for sorting.
#   * Add `/autosort debug` command.
#   * Add ${info:autosort_replace,from,to,text} to replace substrings in sort rules.
#   * Add ${info:autosort_order,value,first,second,third} to ease writing sort rules.
#   * Make tab completion context aware.
# 2.8:
#   * Fix compatibility with python 3 regarding unicode handling.
# 2.7:
#   * Fix sorting of buffers with spaces in their name.
# 2.6:
#   * Ignore case in rules when doing case insensitive sorting.
# 2.5:
#   * Fix handling unicode buffer names.
#   * Add hint to set irc.look.server_buffer to independent and buffers.look.indenting to on.
# 2.4:
#   * Make script python3 compatible.
# 2.3:
#   * Fix sorting items without score last (regressed in 2.2).
# 2.2:
#   * Add configuration option for signals that trigger a sort.
#   * Add command to manually trigger a sort (/autosort sort).
#   * Add replacement patterns to apply before sorting.
# 2.1:
#   * Fix some minor style issues.
# 2.0:
#   * Allow for custom sort rules.
#


import json
import math
import re
import sys
import time
import weechat

SCRIPT_NAME     = 'autosort'
SCRIPT_AUTHOR   = 'Maarten de Vries <maarten@de-vri.es>'
SCRIPT_VERSION  = '3.6'
SCRIPT_LICENSE  = 'GPL3'
SCRIPT_DESC     = 'Flexible automatic (or manual) buffer sorting based on eval expressions.'


config             = None
hooks              = []
signal_delay_timer = None
sort_limit_timer   = None
sort_queued        = False


# Make sure that unicode, bytes and str are always available in python2 and 3.
# For python 2, str == bytes
# For python 3, str == unicode
if sys.version_info[0] >= 3:
	unicode = str

def ensure_str(input):
	'''
	Make sure the given type if the correct string type for the current python version.
	That means bytes for python2 and unicode for python3.
	'''
	if not isinstance(input, str):
		if isinstance(input, bytes):
			return input.encode('utf-8')
		if isinstance(input, unicode):
			return input.decode('utf-8')
	return input


if hasattr(time, 'perf_counter'):
	perf_counter = time.perf_counter
else:
	perf_counter = time.clock

def casefold(string):
	if hasattr(string, 'casefold'): return string.casefold()
	# Fall back to lowercasing for python2.
	return string.lower()

def list_swap(values, a, b):
	values[a], values[b] = values[b], values[a]

def list_move(values, old_index, new_index):
	values.insert(new_index, values.pop(old_index))

def list_find(collection, value):
	for i, elem in enumerate(collection):
		if elem == value: return i
	return None

class HumanReadableError(Exception):
	pass

def parse_int(arg, arg_name = 'argument'):
	''' Parse an integer and provide a more human readable error. '''
	arg = arg.strip()
	try:
		return int(arg)
	except ValueError:
		raise HumanReadableError('Invalid {0}: expected integer, got "{1}".'.format(arg_name, arg))

def decode_rules(blob):
	parsed = json.loads(blob)
	if not isinstance(parsed, list):
		log('Malformed rules, expected a JSON encoded list of strings, but got a {0}. No rules have been loaded. Please fix the setting manually.'.format(type(parsed)))
		return []

	for i, entry in enumerate(parsed):
		if not isinstance(entry, (str, unicode)):
			log('Rule #{0} is not a string but a {1}. No rules have been loaded. Please fix the setting manually.'.format(i, type(entry)))
			return []

	return parsed

def decode_helpers(blob):
	parsed = json.loads(blob)
	if not isinstance(parsed, dict):
		log('Malformed helpers, expected a JSON encoded dictionary but got a {0}. No helpers have been loaded. Please fix the setting manually.'.format(type(parsed)))
		return {}

	for key, value in parsed.items():
		if not isinstance(value, (str, unicode)):
			log('Helper "{0}" is not a string but a {1}. No helpers have been loaded. Please fix setting manually.'.format(key, type(value)))
			return {}
	return parsed

class Config:
	''' The autosort configuration. '''

	default_rules = json.dumps([
		'${core_first}',
		'${irc_last}',
		'${buffer.plugin.name}',
		'${irc_raw_first}',
		'${if:${plugin}==irc?${server}}',
		'${if:${plugin}==irc?${info:autosort_order,${type},server,*,channel,private}}',
		'${if:${plugin}==irc?${hashless_name}}',
		'${buffer.full_name}',
	])

	default_helpers = json.dumps({
		'core_first':     '${if:${buffer.full_name}!=core.weechat}',
		'irc_first':      '${if:${buffer.plugin.name}!=irc}',
		'irc_last':       '${if:${buffer.plugin.name}==irc}',
		'irc_raw_first':  '${if:${buffer.full_name}!=irc.irc_raw}',
		'irc_raw_last':   '${if:${buffer.full_name}==irc.irc_raw}',
		'hashless_name':  '${info:autosort_replace,#,,${info:autosort_escape,${buffer.name}}}',
	})

	default_signal_delay = 5
	default_sort_limit   = 100

	default_signals = 'buffer_opened buffer_merged buffer_unmerged buffer_renamed'

	def __init__(self, filename):
		''' Initialize the configuration. '''

		self.filename         = filename
		self.config_file      = weechat.config_new(self.filename, '', '')
		self.sorting_section  = None
		self.v3_section       = None

		self.case_sensitive   = False
		self.rules            = []
		self.helpers          = {}
		self.signals          = []
		self.signal_delay     = Config.default_signal_delay,
		self.sort_limit       = Config.default_sort_limit,
		self.sort_on_config   = True
		self.debug_log        = False

		self.__case_sensitive = None
		self.__rules          = None
		self.__helpers        = None
		self.__signals        = None
		self.__signal_delay   = None
		self.__sort_limit     = None
		self.__sort_on_config = None
		self.__debug_log      = None

		if not self.config_file:
			log('Failed to initialize configuration file "{0}".'.format(self.filename))
			return

		self.sorting_section = weechat.config_new_section(self.config_file, 'sorting', False, False, '', '', '', '', '', '', '', '', '', '')
		self.v3_section      = weechat.config_new_section(self.config_file, 'v3',      False, False, '', '', '', '', '', '', '', '', '', '')

		if not self.sorting_section:
			log('Failed to initialize section "sorting" of configuration file.')
			weechat.config_free(self.config_file)
			return

		self.__case_sensitive = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'case_sensitive', 'boolean',
			'If this option is on, sorting is case sensitive.',
			'', 0, 0, 'off', 'off', 0,
			'', '', '', '', '', ''
		)

		weechat.config_new_option(
			self.config_file, self.sorting_section,
			'rules', 'string',
			'Sort rules used by autosort v2.x and below. Not used by autosort anymore.',
			'', 0, 0, '', '', 0,
			'', '', '', '', '', ''
		)

		weechat.config_new_option(
			self.config_file, self.sorting_section,
			'replacements', 'string',
			'Replacement patterns used by autosort v2.x and below. Not used by autosort anymore.',
			'', 0, 0, '', '', 0,
			'', '', '', '', '', ''
		)

		self.__rules = weechat.config_new_option(
			self.config_file, self.v3_section,
			'rules', 'string',
			'An ordered list of sorting rules encoded as JSON. See /help autosort for commands to manipulate these rules.',
			'', 0, 0, Config.default_rules, Config.default_rules, 0,
			'', '', '', '', '', ''
		)

		self.__helpers = weechat.config_new_option(
			self.config_file, self.v3_section,
			'helpers', 'string',
			'A dictionary helper variables to use in the sorting rules, encoded as JSON. See /help autosort for commands to manipulate these helpers.',
			'', 0, 0, Config.default_helpers, Config.default_helpers, 0,
			'', '', '', '', '', ''
		)

		self.__signals = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'signals', 'string',
			'A space separated list of signals that will cause autosort to resort your buffer list.',
			'', 0, 0, Config.default_signals, Config.default_signals, 0,
			'', '', '', '', '', ''
		)

		self.__signal_delay = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'signal_delay', 'integer',
			'Delay in milliseconds to wait after a signal before sorting the buffer list. This prevents triggering many times if multiple signals arrive in a short time. It can also be needed to wait for buffer localvars to be available.',
			'', 0, 1000, str(Config.default_signal_delay), str(Config.default_signal_delay), 0,
			'', '', '', '', '', ''
		)

		self.__sort_limit = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'sort_limit', 'integer',
			'Minimum delay in milliseconds to wait after sorting before signals can trigger a sort again. This is effectively a rate limit on sorting. Keeping signal_delay low while setting this higher can reduce excessive sorting without a long initial delay.',
			'', 0, 1000, str(Config.default_sort_limit), str(Config.default_sort_limit), 0,
			'', '', '', '', '', ''
		)

		self.__sort_on_config = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'sort_on_config_change', 'boolean',
			'Decides if the buffer list should be sorted when autosort configuration changes.',
			'', 0, 0, 'on', 'on', 0,
			'', '', '', '', '', ''
		)

		self.__debug_log = weechat.config_new_option(
			self.config_file, self.sorting_section,
			'debug_log', 'boolean',
			'If enabled, print more debug messages. Not recommended for normal usage.',
			'', 0, 0, 'off', 'off', 0,
			'', '', '', '', '', ''
		)

		if weechat.config_read(self.config_file) != weechat.WEECHAT_RC_OK:
			log('Failed to load configuration file.')

		if weechat.config_write(self.config_file) != weechat.WEECHAT_RC_OK:
			log('Failed to write configuration file.')

		self.reload()

	def reload(self):
		''' Load configuration variables. '''

		self.case_sensitive = weechat.config_boolean(self.__case_sensitive)

		rules_blob    = weechat.config_string(self.__rules)
		helpers_blob  = weechat.config_string(self.__helpers)
		signals_blob  = weechat.config_string(self.__signals)

		self.rules          = decode_rules(rules_blob)
		self.helpers        = decode_helpers(helpers_blob)
		self.signals        = signals_blob.split()
		self.signal_delay   = weechat.config_integer(self.__signal_delay)
		self.sort_limit     = weechat.config_integer(self.__sort_limit)
		self.sort_on_config = weechat.config_boolean(self.__sort_on_config)
		self.debug_log      = weechat.config_boolean(self.__debug_log)

	def save_rules(self, run_callback = True):
		''' Save the current rules to the configuration. '''
		weechat.config_option_set(self.__rules, json.dumps(self.rules), run_callback)

	def save_helpers(self, run_callback = True):
		''' Save the current helpers to the configuration. '''
		weechat.config_option_set(self.__helpers, json.dumps(self.helpers), run_callback)


def pad(sequence, length, padding = None):
	''' Pad a list until is has a certain length. '''
	return sequence + [padding] * max(0, (length - len(sequence)))

def log(message, buffer = 'NULL'):
	weechat.prnt(buffer, 'autosort: {0}'.format(message))

def debug(message, buffer = 'NULL'):
	if config.debug_log:
		weechat.prnt(buffer, 'autosort: debug: {0}'.format(message))

def get_buffers():
	''' Get a list of all the buffers in weechat. '''
	hdata  = weechat.hdata_get('buffer')
	buffer = weechat.hdata_get_list(hdata, "gui_buffers");

	result = []
	while buffer:
		number = weechat.hdata_integer(hdata, buffer, 'number')
		result.append((number, buffer))
		buffer = weechat.hdata_pointer(hdata, buffer, 'next_buffer')
	return hdata, result

class MergedBuffers(list):
	""" A list of merged buffers, possibly of size 1. """
	def __init__(self, number):
		super(MergedBuffers, self).__init__()
		self.number = number

def merge_buffer_list(buffers):
	'''
	Group merged buffers together.
	The output is a list of MergedBuffers.
	'''
	if not buffers: return []
	result = {}
	for number, buffer in buffers:
		if number not in result: result[number] = MergedBuffers(number)
		result[number].append(buffer)
	return result.values()

def sort_buffers(hdata, buffers, rules, helpers, case_sensitive):
	for merged in buffers:
		for buffer in merged:
			name = weechat.hdata_string(hdata, buffer, 'name')

	return sorted(buffers, key=merged_sort_key(rules, helpers, case_sensitive))

def buffer_sort_key(rules, helpers, case_sensitive):
	''' Create a sort key function for a list of lists of merged buffers. '''
	def key(buffer):
		extra_vars = {}
		for helper_name, helper in sorted(helpers.items()):
			expanded = weechat.string_eval_expression(helper, {"buffer": buffer}, {}, {})
			extra_vars[helper_name] = expanded if case_sensitive else casefold(expanded)
		result = []
		for rule in rules:
			expanded = weechat.string_eval_expression(rule, {"buffer": buffer}, extra_vars, {})
			result.append(expanded if case_sensitive else casefold(expanded))
		return result

	return key

def merged_sort_key(rules, helpers, case_sensitive):
	buffer_key = buffer_sort_key(rules, helpers, case_sensitive)
	def key(merged):
		best = None
		for buffer in merged:
			this = buffer_key(buffer)
			if best is None or this < best: best = this
		return best
	return key

def apply_buffer_order(buffers):
	''' Sort the buffers in weechat according to the given order. '''
	for i, buffer in enumerate(buffers):
		weechat.buffer_set(buffer[0], "number", str(i + 1))

def split_args(args, expected, optional = 0):
	''' Split an argument string in the desired number of arguments. '''
	split = args.split(' ', expected - 1)
	if (len(split) < expected):
		raise HumanReadableError('Expected at least {0} arguments, got {1}.'.format(expected, len(split)))
	return split[:-1] + pad(split[-1].split(' ', optional), optional + 1, '')

def do_sort(verbose = False):
	start = perf_counter()

	hdata, buffers = get_buffers()
	buffers = merge_buffer_list(buffers)
	buffers = sort_buffers(hdata, buffers, config.rules, config.helpers, config.case_sensitive)
	apply_buffer_order(buffers)

	elapsed = perf_counter() - start
	if verbose:
		log("Finished sorting buffers in {0:.4f} seconds.".format(elapsed))
	else:
		debug("Finished sorting buffers in {0:.4f} seconds.".format(elapsed))

def command_sort(buffer, command, args):
	''' Sort the buffers and print a confirmation. '''
	do_sort(True)
	return weechat.WEECHAT_RC_OK

def command_debug(buffer, command, args):
	hdata, buffers = get_buffers()
	buffers = merge_buffer_list(buffers)

	# Show evaluation results.
	log('Individual evaluation results:')
	start = perf_counter()
	key = buffer_sort_key(config.rules, config.helpers, config.case_sensitive)
	results = []
	for merged in buffers:
		for buffer in merged:
			fullname = weechat.hdata_string(hdata, buffer, 'full_name')
			results.append((fullname, key(buffer)))
	elapsed = perf_counter() - start

	for fullname, result in results:
		fullname = ensure_str(fullname)
		result = [ensure_str(x) for x in result]
		log('{0}: {1}'.format(fullname, result))
	log('Computing evaluation results took {0:.4f} seconds.'.format(elapsed))

	return weechat.WEECHAT_RC_OK

def command_rule_list(buffer, command, args):
	''' Show the list of sorting rules. '''
	output = 'Sorting rules:\n'
	for i, rule in enumerate(config.rules):
		output += '    {0}: {1}\n'.format(i, rule)
	if not len(config.rules):
		output += '    No sorting rules configured.\n'
	log(output )

	return weechat.WEECHAT_RC_OK


def command_rule_add(buffer, command, args):
	''' Add a rule to the rule list. '''
	config.rules.append(args)
	config.save_rules()
	command_rule_list(buffer, command, '')

	return weechat.WEECHAT_RC_OK


def command_rule_insert(buffer, command, args):
	''' Insert a rule at the desired position in the rule list. '''
	index, rule = split_args(args, 2)
	index = parse_int(index, 'index')

	config.rules.insert(index, rule)
	config.save_rules()
	command_rule_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_rule_update(buffer, command, args):
	''' Update a rule in the rule list. '''
	index, rule = split_args(args, 2)
	index = parse_int(index, 'index')

	config.rules[index] = rule
	config.save_rules()
	command_rule_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_rule_delete(buffer, command, args):
	''' Delete a rule from the rule list. '''
	index = args.strip()
	index = parse_int(index, 'index')

	config.rules.pop(index)
	config.save_rules()
	command_rule_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_rule_move(buffer, command, args):
	''' Move a rule to a new position. '''
	index_a, index_b = split_args(args, 2)
	index_a = parse_int(index_a, 'index')
	index_b = parse_int(index_b, 'index')

	list_move(config.rules, index_a, index_b)
	config.save_rules()
	command_rule_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_rule_swap(buffer, command, args):
	''' Swap two rules. '''
	index_a, index_b = split_args(args, 2)
	index_a = parse_int(index_a, 'index')
	index_b = parse_int(index_b, 'index')

	list_swap(config.rules, index_a, index_b)
	config.save_rules()
	command_rule_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_helper_list(buffer, command, args):
	''' Show the list of helpers. '''
	output = 'Helper variables:\n'

	width = max(map(lambda x: len(x) if len(x) <= 30 else 0, config.helpers.keys()))

	for name, expression in sorted(config.helpers.items()):
		output += '    {0:>{width}}: {1}\n'.format(name, expression, width=width)
	if not len(config.helpers):
		output += '    No helper variables configured.'
	log(output)

	return weechat.WEECHAT_RC_OK


def command_helper_set(buffer, command, args):
	''' Add/update a helper to the helper list. '''
	name, expression = split_args(args, 2)

	config.helpers[name] = expression
	config.save_helpers()
	command_helper_list(buffer, command, '')

	return weechat.WEECHAT_RC_OK

def command_helper_delete(buffer, command, args):
	''' Delete a helper from the helper list. '''
	name = args.strip()

	del config.helpers[name]
	config.save_helpers()
	command_helper_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_helper_rename(buffer, command, args):
	''' Rename a helper to a new position. '''
	old_name, new_name = split_args(args, 2)

	try:
		config.helpers[new_name] = config.helpers[old_name]
		del config.helpers[old_name]
	except KeyError:
		raise HumanReadableError('No such helper: {0}'.format(old_name))
	config.save_helpers()
	command_helper_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK


def command_helper_swap(buffer, command, args):
	''' Swap two helpers. '''
	a, b = split_args(args, 2)
	try:
		config.helpers[b], config.helpers[a] = config.helpers[a], config.helpers[b]
	except KeyError as e:
		raise HumanReadableError('No such helper: {0}'.format(e.args[0]))

	config.helpers.swap(index_a, index_b)
	config.save_helpers()
	command_helper_list(buffer, command, '')
	return weechat.WEECHAT_RC_OK

def call_command(buffer, command, args, subcommands):
	''' Call a subcommand from a dictionary. '''
	subcommand, tail = pad(args.split(' ', 1), 2, '')
	subcommand = subcommand.strip()
	if (subcommand == ''):
		child   = subcommands.get(' ')
	else:
		command = command + [subcommand]
		child   = subcommands.get(subcommand)

	if isinstance(child, dict):
		return call_command(buffer, command, tail, child)
	elif callable(child):
		return child(buffer, command, tail)

	log('{0}: command not found'.format(' '.join(command)))
	return weechat.WEECHAT_RC_ERROR

def on_signal(data, signal, signal_data):
	global signal_delay_timer
	global sort_queued

	# If the sort limit timeout is started, we're in the hold-off time after sorting, just queue a sort.
	if sort_limit_timer is not None:
		if sort_queued:
			debug('Signal {0} ignored, sort limit timeout is active and sort is already queued.'.format(signal))
		else:
			debug('Signal {0} received but sort limit timeout is active, sort is now queued.'.format(signal))
		sort_queued = True
		return weechat.WEECHAT_RC_OK

	# If the signal delay timeout is started, a signal was recently received, so ignore this signal.
	if signal_delay_timer is not None:
		debug('Signal {0} ignored, signal delay timeout active.'.format(signal))
		return weechat.WEECHAT_RC_OK

	# Otherwise, start the signal delay timeout.
	debug('Signal {0} received, starting signal delay timeout of {1} ms.'.format(signal, config.signal_delay))
	weechat.hook_timer(config.signal_delay, 0, 1, "on_signal_delay_timeout", "")
	return weechat.WEECHAT_RC_OK

def on_signal_delay_timeout(pointer, remaining_calls):
	""" Called when the signal_delay_timer triggers. """
	global signal_delay_timer
	global sort_limit_timer
	global sort_queued

	signal_delay_timer = None

	# If the sort limit timeout was started, we're still in the no-sort period, so just queue a sort.
	if sort_limit_timer is not None:
		debug('Signal delay timeout expired, but sort limit timeout is active, sort is now queued.')
		sort_queued = True
		return weechat.WEECHAT_RC_OK

	# Time to sort!
	debug('Signal delay timeout expired, starting sort.')
	do_sort()

	# Start the sort limit timeout if not disabled.
	if config.sort_limit > 0:
		debug('Starting sort limit timeout of {0} ms.'.format(config.sort_limit))
		sort_limit_timer = weechat.hook_timer(config.sort_limit, 0, 1, "on_sort_limit_timeout", "")

	return weechat.WEECHAT_RC_OK

def on_sort_limit_timeout(pointer, remainin_calls):
	""" Called when de sort_limit_timer triggers. """
	global sort_limit_timer
	global sort_queued

	# If no signal was received during the timeout, we're done.
	if not sort_queued:
		debug('Sort limit timeout expired without receiving a signal.')
		sort_limit_timer = None
		return weechat.WEECHAT_RC_OK

	# Otherwise it's time to sort.
	debug('Signal received during sort limit timeout, starting queued sort.')
	do_sort()
	sort_queued = False

	# Start the sort limit timeout again if not disabled.
	if config.sort_limit > 0:
		debug('Starting sort limit timeout of {0} ms.'.format(config.sort_limit))
		sort_limit_timer = weechat.hook_timer(config.sort_limit, 0, 1, "on_sort_limit_timeout", "")

	return weechat.WEECHAT_RC_OK


def apply_config():
	# Unhook all signals and hook the new ones.
	for hook in hooks:
		weechat.unhook(hook)
	for signal in config.signals:
		hooks.append(weechat.hook_signal(signal, 'on_signal', ''))

	if config.sort_on_config:
		debug('Sorting because configuration changed.')
		do_sort()

def on_config_changed(*args, **kwargs):
	''' Called whenever the configuration changes. '''
	config.reload()
	apply_config()

	return weechat.WEECHAT_RC_OK

def parse_arg(args):
	if not args: return '', None

	result  = ''
	escaped = False
	for i, c in enumerate(args):
		if not escaped:
			if c == '\\':
				escaped = True
				continue
			elif c == ',':
				return result, args[i+1:]
		result  += c
		escaped  = False
	return result, None

def parse_args(args, max = None):
	result = []
	i = 0
	while max is None or i < max:
		i += 1
		arg, args = parse_arg(args)
		if arg is None: break
		result.append(arg)
		if args is None: break
	return result, args

def on_info_escape(pointer, name, arguments):
	result = ''
	for c in arguments:
		if c == '\\':
			result += '\\\\'
		elif c == ',':
			result += '\\,'
		else:
			result +=c
	return result

def on_info_replace(pointer, name, arguments):
	arguments, rest = parse_args(arguments, 3)
	if rest or len(arguments) < 3:
		log('usage: ${{info:{0},old,new,text}}'.format(name))
		return ''
	old, new, text = arguments

	return text.replace(old, new)

def on_info_order(pointer, name, arguments):
	arguments, rest = parse_args(arguments)
	if len(arguments) < 1:
		log('usage: ${{info:{0},value,first,second,third,...}}'.format(name))
		return ''

	value = arguments[0]
	keys  = arguments[1:]
	if not keys: return '0'

	# Find the value in the keys (or '*' if we can't find it)
	result = list_find(keys, value)
	if result is None: result = list_find(keys, '*')
	if result is None: result = len(keys)

	# Pad result with leading zero to make sure string sorting works.
	width = int(math.log10(len(keys))) + 1
	return '{0:0{1}}'.format(result, width)


def on_autosort_command(data, buffer, args):
	''' Called when the autosort command is invoked. '''
	try:
		return call_command(buffer, ['/autosort'], args, {
			' ':      command_sort,
			'sort':   command_sort,
			'debug':  command_debug,

			'rules': {
				' ':         command_rule_list,
				'list':      command_rule_list,
				'add':       command_rule_add,
				'insert':    command_rule_insert,
				'update':    command_rule_update,
				'delete':    command_rule_delete,
				'move':      command_rule_move,
				'swap':      command_rule_swap,
			},
			'helpers': {
				' ':      command_helper_list,
				'list':   command_helper_list,
				'set':    command_helper_set,
				'delete': command_helper_delete,
				'rename': command_helper_rename,
				'swap':   command_helper_swap,
			},
		})
	except HumanReadableError as e:
		log(e)
		return weechat.WEECHAT_RC_ERROR

def add_completions(completion, words):
	for word in words:
		weechat.hook_completion_list_add(completion, word, 0, weechat.WEECHAT_LIST_POS_END)

def autosort_complete_rules(words, completion):
	if len(words) == 0:
		add_completions(completion, ['add', 'delete', 'insert', 'list', 'move', 'swap', 'update'])
	if len(words) == 1 and words[0] in ('delete', 'insert', 'move', 'swap', 'update'):
		add_completions(completion, map(str, range(len(config.rules))))
	if len(words) == 2 and words[0] in ('move', 'swap'):
		add_completions(completion, map(str, range(len(config.rules))))
	if len(words) == 2 and words[0] in ('update'):
		try:
			add_completions(completion, [config.rules[int(words[1])]])
		except KeyError: pass
		except ValueError: pass
	else:
		add_completions(completion, [''])
	return weechat.WEECHAT_RC_OK

def autosort_complete_helpers(words, completion):
	if len(words) == 0:
		add_completions(completion, ['delete', 'list', 'rename', 'set', 'swap'])
	elif len(words) == 1 and words[0] in ('delete', 'rename', 'set', 'swap'):
		add_completions(completion, sorted(config.helpers.keys()))
	elif len(words) == 2 and words[0] == 'swap':
		add_completions(completion, sorted(config.helpers.keys()))
	elif len(words) == 2 and words[0] == 'rename':
		add_completions(completion, sorted(config.helpers.keys()))
	elif len(words) == 2 and words[0] == 'set':
		try:
			add_completions(completion, [config.helpers[words[1]]])
		except KeyError: pass
	return weechat.WEECHAT_RC_OK

def on_autosort_complete(data, name, buffer, completion):
	cmdline = weechat.buffer_get_string(buffer, "input")
	cursor  = weechat.buffer_get_integer(buffer, "input_pos")
	prefix  = cmdline[:cursor]
	words   = prefix.split()[1:]

	# If the current word isn't finished yet,
	# ignore it for coming up with completion suggestions.
	if prefix[-1] != ' ': words = words[:-1]

	if len(words) == 0:
		add_completions(completion, ['debug', 'helpers', 'rules', 'sort'])
	elif words[0] == 'rules':
		return autosort_complete_rules(words[1:], completion)
	elif words[0] == 'helpers':
		return autosort_complete_helpers(words[1:], completion)
	return weechat.WEECHAT_RC_OK

command_description = r'''{*white}# General commands{reset}

{*white}/autosort {brown}sort{reset}
Manually trigger the buffer sorting.

{*white}/autosort {brown}debug{reset}
Show the evaluation results of the sort rules for each buffer.


{*white}# Sorting rule commands{reset}

{*white}/autosort{brown} rules list{reset}
Print the list of sort rules.

{*white}/autosort {brown}rules add {cyan}<expression>{reset}
Add a new rule at the end of the list.

{*white}/autosort {brown}rules insert {cyan}<index> <expression>{reset}
Insert a new rule at the given index in the list.

{*white}/autosort {brown}rules update {cyan}<index> <expression>{reset}
Update a rule in the list with a new expression.

{*white}/autosort {brown}rules delete {cyan}<index>
Delete a rule from the list.

{*white}/autosort {brown}rules move {cyan}<index_from> <index_to>{reset}
Move a rule from one position in the list to another.

{*white}/autosort {brown}rules swap {cyan}<index_a> <index_b>{reset}
Swap two rules in the list


{*white}# Helper variable commands{reset}

{*white}/autosort {brown}helpers list
Print the list of helper variables.

{*white}/autosort {brown}helpers set {cyan}<name> <expression>
Add or update a helper variable with the given name.

{*white}/autosort {brown}helpers delete {cyan}<name>
Delete a helper variable.

{*white}/autosort {brown}helpers rename {cyan}<old_name> <new_name>
Rename a helper variable.

{*white}/autosort {brown}helpers swap {cyan}<name_a> <name_b>
Swap the expressions of two helper variables in the list.


{*white}# Info hooks{reset}
Autosort comes with a number of info hooks to add some extra functionality to regular weechat eval strings.
Info hooks can be used in eval strings in the form of {cyan}${{info:some_hook,arguments}}{reset}.

Commas and backslashes in arguments to autosort info hooks (except for {cyan}${{info:autosort_escape}}{reset}) must be escaped with a backslash.

{*white}${{info:{brown}autosort_replace{white},{cyan}pattern{white},{cyan}replacement{white},{cyan}source{white}}}{reset}
Replace all occurrences of {cyan}pattern{reset} with {cyan}replacement{reset} in the string {cyan}source{reset}.
Can be used to ignore certain strings when sorting by replacing them with an empty string.

For example: {cyan}${{info:autosort_replace,cat,dog,the dog is meowing}}{reset} expands to "the cat is meowing".

{*white}${{info:{brown}autosort_order{white},{cyan}value{white},{cyan}option0{white},{cyan}option1{white},{cyan}option2{white},{cyan}...{white}}}
Generate a zero-padded number that corresponds to the index of {cyan}value{reset} in the list of options.
If one of the options is the special value {brown}*{reset}, then any value not explicitly mentioned will be sorted at that position.
Otherwise, any value that does not match an option is assigned the highest number available.
Can be used to easily sort buffers based on a manual sequence.

For example: {cyan}${{info:autosort_order,${{server}},freenode,oftc,efnet}}{reset} will sort freenode before oftc, followed by efnet and then any remaining servers.
Alternatively, {cyan}${{info:autosort_order,${{server}},freenode,oftc,*,efnet}}{reset} will sort any unlisted servers after freenode and oftc, but before efnet.

{*white}${{info:{brown}autosort_escape{white},{cyan}text{white}}}{reset}
Escape commas and backslashes in {cyan}text{reset} by prepending them with a backslash.
This is mainly useful to pass arbitrary eval strings as arguments to other autosort info hooks.
Otherwise, an eval string that expands to something with a comma would be interpreted as multiple arguments.

For example, it can be used to safely pass buffer names to {cyan}${{info:autosort_replace}}{reset} like so:
{cyan}${{info:autosort_replace,##,#,${{info:autosort_escape,${{buffer.name}}}}}}{reset}.


{*white}# Description
Autosort is a weechat script to automatically keep your buffers sorted. The sort
order can be customized by defining your own sort rules, but the default should
be sane enough for most people. It can also group IRC channel/private buffers
under their server buffer if you like.

{*white}# Sort rules{reset}
Autosort evaluates a list of eval expressions (see {*default}/help eval{reset}) and sorts the
buffers based on evaluated result. Earlier rules will be considered first. Only
if earlier rules produced identical results is the result of the next rule
considered for sorting purposes.

You can debug your sort rules with the `{*default}/autosort debug{reset}` command, which will
print the evaluation results of each rule for each buffer.

{*brown}NOTE:{reset} The sort rules for version 3 are not compatible with version 2 or vice
versa. You will have to manually port your old rules to version 3 if you have any.

{*white}# Helper variables{reset}
You may define helper variables for the main sort rules to keep your rules
readable. They can be used in the main sort rules as variables. For example,
a helper variable named `{cyan}foo{reset}` can be accessed in a main rule with the
string `{cyan}${{foo}}{reset}`.

{*white}# Automatic or manual sorting{reset}
By default, autosort will automatically sort your buffer list whenever a buffer
is opened, merged, unmerged or renamed. This should keep your buffers sorted in
almost all situations. However, you may wish to change the list of signals that
cause your buffer list to be sorted. Simply edit the `{cyan}autosort.sorting.signals{reset}`
option to add or remove any signal you like.

If you remove all signals you can still sort your buffers manually with the
`{*default}/autosort sort{reset}` command. To prevent all automatic sorting, the option
`{cyan}autosort.sorting.sort_on_config_change{reset}` should also be disabled.

{*white}# Recommended settings
For the best visual effect, consider setting the following options:
  {*white}/set {cyan}irc.look.server_buffer{reset} {brown}independent{reset}
  {*white}/set {cyan}buffers.look.indenting{reset} {brown}on{reset}

The first setting allows server buffers to be sorted independently, which is
needed to create a hierarchical tree view of the server and channel buffers.
The second one indents channel and private buffers in the buffer list of the
`{*default}buffers.pl{reset}` script.

If you are using the {*default}buflist{reset} plugin you can (ab)use Unicode to draw a tree
structure with the following setting (modify to suit your need):
  {*white}/set {cyan}buflist.format.indent {brown}"${{color:237}}${{if:${{buffer.next_buffer.local_variables.type}}=~^(channel|private)$?├─:└─}}"{reset}
'''

command_completion = '%(plugin_autosort) %(plugin_autosort) %(plugin_autosort) %(plugin_autosort) %(plugin_autosort)'

info_replace_description = (
	'Replace all occurrences of `pattern` with `replacement` in the string `source`. '
	'Can be used to ignore certain strings when sorting by replacing them with an empty string. '
	'See /help autosort for examples.'
)
info_replace_arguments = 'pattern,replacement,source'

info_order_description = (
	'Generate a zero-padded number that corresponds to the index of `value` in the list of options. '
	'If one of the options is the special value `*`, then any value not explicitly mentioned will be sorted at that position. '
	'Otherwise, any value that does not match an option is assigned the highest number available. '
	'Can be used to easily sort buffers based on a manual sequence. '
	'See /help autosort for examples.'
)
info_order_arguments = 'value,first,second,third,...'

info_escape_description = (
	'Escape commas and backslashes in `text` by prepending them with a backslash. '
	'This is mainly useful to pass arbitrary eval strings as arguments to other autosort info hooks. '
	'Otherwise, an eval string that expands to something with a comma would be interpreted as multiple arguments.'
	'See /help autosort for examples.'
)
info_escape_arguments = 'text'


if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""):
	config = Config('autosort')

	colors = {
		'default':  weechat.color('default'),
		'reset':    weechat.color('reset'),
		'black':    weechat.color('black'),
		'red':      weechat.color('red'),
		'green':    weechat.color('green'),
		'brown':    weechat.color('brown'),
		'yellow':   weechat.color('yellow'),
		'blue':     weechat.color('blue'),
		'magenta':  weechat.color('magenta'),
		'cyan':     weechat.color('cyan'),
		'white':    weechat.color('white'),
		'*default': weechat.color('*default'),
		'*black':   weechat.color('*black'),
		'*red':     weechat.color('*red'),
		'*green':   weechat.color('*green'),
		'*brown':   weechat.color('*brown'),
		'*yellow':  weechat.color('*yellow'),
		'*blue':    weechat.color('*blue'),
		'*magenta': weechat.color('*magenta'),
		'*cyan':    weechat.color('*cyan'),
		'*white':   weechat.color('*white'),
	}

	weechat.hook_config('autosort.*', 'on_config_changed',  '')
	weechat.hook_completion('plugin_autosort', '', 'on_autosort_complete', '')
	weechat.hook_command('autosort', command_description.format(**colors), '', '', command_completion, 'on_autosort_command', '')
	weechat.hook_info('autosort_escape',  info_escape_description,  info_escape_arguments,  'on_info_escape', '')
	weechat.hook_info('autosort_replace', info_replace_description, info_replace_arguments, 'on_info_replace', '')
	weechat.hook_info('autosort_order',   info_order_description,   info_order_arguments,   'on_info_order',   '')

	apply_config()



# -*- coding: utf-8 -*-
#
# Copyright (C) 2013-2014 Germain Z. <germanosz@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

#
# Add vi/vim-like modes to WeeChat.
#


import csv
import json
import os
import re
import subprocess
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO
import time

import weechat


# Script info.
# ============

SCRIPT_NAME = "vimode"
SCRIPT_AUTHOR = "GermainZ <germanosz@gmail.com>"
SCRIPT_VERSION = "0.7"
SCRIPT_LICENSE = "GPL3"
SCRIPT_DESC = ("Add vi/vim-like modes and keybindings to WeeChat.")


# Global variables.
# =================

# General.
# --------

# Halp! Halp! Halp!
GITHUB_BASE = "https://github.com/GermainZ/weechat-vimode/blob/master/"
README_URL = GITHUB_BASE + "README.md"
FAQ_KEYBINDINGS = GITHUB_BASE + "FAQ.md#problematic-key-bindings"
FAQ_ESC = GITHUB_BASE + "FAQ.md#esc-key-not-being-detected-instantly"

# Holds the text of the tab-completions for the command-line mode.
cmd_compl_text = ""
# Holds the original text of the command-line mode, used for completion.
cmd_text_orig = None
# Index of current suggestion, used for completion.
cmd_compl_pos = 0
# Used for command-line mode history.
cmd_history = []
cmd_history_index = 0
# Used to store the content of the input line when going into COMMAND mode.
input_line_backup = {}
# Mode we're in. One of INSERT, NORMAL, REPLACE, COMMAND or SEARCH.
# SEARCH is only used if search_vim is enabled.
mode = "INSERT"
# Holds normal commands (e.g. "dd").
vi_buffer = ""
# See `cb_key_combo_default()`.
esc_pressed = 0
# See `cb_key_pressed()`.
last_signal_time = 0
# See `start_catching_keys()` for more info.
catching_keys_data = {'amount': 0}
# Used for ; and , to store the last f/F/t/T motion.
last_search_motion = {'motion': None, 'data': None}
# Used for undo history.
undo_history = {}
undo_history_index = {}
# Holds mode colors (loaded from vimode_settings).
mode_colors = {}

# Script options.
vimode_settings = {
    'no_warn': ("off", ("don't warn about problematic keybindings and "
                        "tmux/screen")),
    'copy_clipboard_cmd': ("xclip -selection c",
                           ("command used to copy to clipboard; must read "
                            "input from stdin")),
    'paste_clipboard_cmd': ("xclip -selection c -o",
                            ("command used to paste clipboard; must output "
                             "content to stdout")),
    'imap_esc': ("", ("use alternate mapping to enter Normal mode while in "
                      "Insert mode; having it set to 'jk' is similar to "
                      "`:imap jk <Esc>` in vim")),
    'imap_esc_timeout': ("1000", ("time in ms to wait for the imap_esc "
                                  "sequence to complete")),
    'search_vim': ("off", ("allow n/N usage after searching (requires an extra"
                           " <Enter> to return to normal mode)")),
    'user_mappings': ("", ("see the `:nmap` command in the README for more "
                           "info; please do not modify this field manually "
                           "unless you know what you're doing")),
    'mode_indicator_prefix': ("", "prefix for the bar item mode_indicator"),
    'mode_indicator_suffix': ("", "suffix for the bar item mode_indicator"),
    'mode_indicator_normal_color': ("white",
                                    "color for mode indicator in Normal mode"),
    'mode_indicator_normal_color_bg': ("gray",
                                       ("background color for mode indicator "
                                        "in Normal mode")),
    'mode_indicator_insert_color': ("white",
                                    "color for mode indicator in Insert mode"),
    'mode_indicator_insert_color_bg': ("blue",
                                       ("background color for mode indicator "
                                        "in Insert mode")),
    'mode_indicator_replace_color': ("white",
                                     "color for mode indicator in Replace mode"),
    'mode_indicator_replace_color_bg': ("red",
                                        ("background color for mode indicator "
                                         "in Replace mode")),
    'mode_indicator_cmd_color': ("white",
                                 "color for mode indicator in Command mode"),
    'mode_indicator_cmd_color_bg': ("cyan",
                                    ("background color for mode indicator in "
                                     "Command mode")),
    'mode_indicator_search_color': ("white",
                                    "color for mode indicator in Search mode"),
    'mode_indicator_search_color_bg': ("magenta",
                                       ("background color for mode indicator "
                                        "in Search mode")),
    'line_number_prefix': ("", "prefix for line numbers"),
    'line_number_suffix': (" ", "suffix for line numbers")
}


# Regex patterns.
# ---------------

WHITESPACE = re.compile(r"\s")
IS_KEYWORD = re.compile(r"[a-zA-Z0-9_@À-ÿ]")
REGEX_MOTION_LOWERCASE_W = re.compile(r"\b\S|(?<=\s)\S")
REGEX_MOTION_UPPERCASE_W = re.compile(r"(?<=\s)\S")
REGEX_MOTION_UPPERCASE_E = re.compile(r"\S(?!\S)")
REGEX_MOTION_UPPERCASE_B = REGEX_MOTION_UPPERCASE_E
REGEX_MOTION_G_UPPERCASE_E = REGEX_MOTION_UPPERCASE_W
REGEX_MOTION_CARRET = re.compile(r"\S")
REGEX_INT = r"[0-9]"
REGEX_MAP_KEYS_1 = {
    re.compile("<([^>]*-)Left>", re.IGNORECASE): '<\\1\x01[[D>',
    re.compile("<([^>]*-)Right>", re.IGNORECASE): '<\\1\x01[[C>',
    re.compile("<([^>]*-)Up>", re.IGNORECASE): '<\\1\x01[[A>',
    re.compile("<([^>]*-)Down>", re.IGNORECASE): '<\\1\x01[[B>',
    re.compile("<Left>", re.IGNORECASE): '\x01[[D',
    re.compile("<Right>", re.IGNORECASE): '\x01[[C',
    re.compile("<Up>", re.IGNORECASE): '\x01[[A',
    re.compile("<Down>", re.IGNORECASE): '\x01[[B'
}
REGEX_MAP_KEYS_2 = {
    re.compile(r"<C-([^>]*)>", re.IGNORECASE): '\x01\\1',
    re.compile(r"<M-([^>]*)>", re.IGNORECASE): '\x01[\\1'
}

# Regex used to detect problematic keybindings.
# For example: meta-wmeta-s is bound by default to ``/window swap``.
#    If the user pressed Esc-w, WeeChat will detect it as meta-w and will not
#    send any signal to `cb_key_combo_default()` just yet, since it's the
#    beginning of a known key combo.
#    Instead, `cb_key_combo_default()` will receive the Esc-ws signal, which
#    becomes "ws" after removing the Esc part, and won't know how to handle it.
REGEX_PROBLEMATIC_KEYBINDINGS = re.compile(r"meta-\w(meta|ctrl)")


# Vi commands.
# ------------

def cmd_nmap(args):
    """Add a user-defined key mapping.

    Some (but not all) vim-like key codes are supported to simplify things for
    the user: <Up>, <Down>, <Left>, <Right>, <C-...> and <M-...>.

    See Also:
        `cmd_unmap()`.
    """
    args = args.strip()
    if not args:
        mappings = vimode_settings['user_mappings']
        if mappings:
            weechat.prnt("", "User-defined key mappings:")
            for key, mapping in mappings.items():
                weechat.prnt("", "{} -> {}".format(key, mapping))
        else:
            weechat.prnt("", "nmap: no mapping found.")
    elif not " " in args:
        weechat.prnt("", "nmap syntax -> :nmap {lhs} {rhs}")
    else:
        key, mapping = args.split(" ", 1)
        # First pass of replacements. We perform two passes as a simple way to
        # avoid incorrect replacements due to dictionaries not being
        # insertion-ordered prior to Python 3.7.
        for regex, repl in REGEX_MAP_KEYS_1.items():
            key = regex.sub(repl, key)
            mapping = regex.sub(repl, mapping)
        # Second pass of replacements.
        for regex, repl in REGEX_MAP_KEYS_2.items():
            key = regex.sub(repl, key)
            mapping = regex.sub(repl, mapping)
        mappings = vimode_settings['user_mappings']
        mappings[key] = mapping
        weechat.config_set_plugin('user_mappings', json.dumps(mappings))
        vimode_settings['user_mappings'] = mappings

def cmd_nunmap(args):
    """Remove a user-defined key mapping.

    See Also:
        `cmd_map()`.
    """
    args = args.strip()
    if not args:
        weechat.prnt("", "nunmap syntax -> :unmap {lhs}")
    else:
        key = args
        for regex, repl in REGEX_MAP_KEYS_1.items():
            key = regex.sub(repl, key)
        for regex, repl in REGEX_MAP_KEYS_2.items():
            key = regex.sub(repl, key)
        mappings = vimode_settings['user_mappings']
        if key in mappings:
            del mappings[key]
            weechat.config_set_plugin('user_mappings', json.dumps(mappings))
            vimode_settings['user_mappings'] = mappings
        else:
            weechat.prnt("", "nunmap: No such mapping")

# See Also: `cb_exec_cmd()`.
VI_COMMAND_GROUPS = {('h', 'help'): "/help",
                     ('qa', 'qall', 'quita', 'quitall'): "/exit",
                     ('q', 'quit'): "/close",
                     ('w', 'write'): "/save",
                     ('bN', 'bNext', 'bp', 'bprevious'): "/buffer -1",
                     ('bn', 'bnext'): "/buffer +1",
                     ('bd', 'bdel', 'bdelete'): "/close",
                     ('b#',): "/input jump_last_buffer_displayed",
                     ('b', 'bu', 'buf', 'buffer'): "/buffer",
                     ('sp', 'split'): "/window splith",
                     ('vs', 'vsplit'): "/window splitv",
                     ('nm', 'nmap'): cmd_nmap,
                     ('nun', 'nunmap'): cmd_nunmap}

VI_COMMANDS = dict()
for T, v in VI_COMMAND_GROUPS.items():
    VI_COMMANDS.update(dict.fromkeys(T, v))


# Vi operators.
# -------------

# Each operator must have a corresponding function, called "operator_X" where
# X is the operator. For example: `operator_c()`.
VI_OPERATORS = ["c", "d", "y"]


# Vi motions.
# -----------

# Vi motions. Each motion must have a corresponding function, called
# "motion_X" where X is the motion (e.g. `motion_w()`).
# See Also: `SPECIAL_CHARS`.
VI_MOTIONS = ["w", "e", "b", "^", "$", "h", "l", "W", "E", "B", "f", "F", "t",
              "T", "ge", "gE", "0"]

# Special characters for motions. The corresponding function's name is
# converted before calling. For example, "^" will call `motion_carret` instead
# of `motion_^` (which isn't allowed because of illegal characters).
SPECIAL_CHARS = {'^': "carret",
                 '$': "dollar"}


# Methods for vi operators, motions and key bindings.
# ===================================================

# Documented base examples:
# -------------------------

def operator_base(buf, input_line, pos1, pos2, overwrite):
    """Operator method example.

    Args:
        buf (str): pointer to the current WeeChat buffer.
        input_line (str): the content of the input line.
        pos1 (int): the starting position of the motion.
        pos2 (int): the ending position of the motion.
        overwrite (bool, optional): whether the character at the cursor's new
            position should be overwritten or not (for inclusive motions).
            Defaults to False.

    Notes:
        Should be called "operator_X", where X is the operator, and defined in
        `VI_OPERATORS`.
        Must perform actions (e.g. modifying the input line) on its own,
        using the WeeChat API.

    See Also:
        For additional examples, see `operator_d()` and
        `operator_y()`.
    """
    # Get start and end positions.
    start = min(pos1, pos2)
    end = max(pos1, pos2)
    # Print the text the operator should go over.
    weechat.prnt("", "Selection: %s" % input_line[start:end])

def motion_base(input_line, cur, count):
    """Motion method example.

    Args:
        input_line (str): the content of the input line.
        cur (int): the position of the cursor.
        count (int): the amount of times to multiply or iterate the action.

    Returns:
        A tuple containing three values:
            int: the new position of the cursor.
            bool: True if the motion is inclusive, False otherwise.
            bool: True if the motion is catching, False otherwise.
                See `start_catching_keys()` for more info on catching motions.

    Notes:
        Should be called "motion_X", where X is the motion, and defined in
        `VI_MOTIONS`.
        Must not modify the input line directly.

    See Also:
        For additional examples, see `motion_w()` (normal motion) and
        `motion_f()` (catching motion).
    """
    # Find (relative to cur) position of next number.
    pos = get_pos(input_line, REGEX_INT, cur, True, count)
    # Return the new (absolute) cursor position.
    # This motion is exclusive, so overwrite is False.
    return cur + pos, False

def key_base(buf, input_line, cur, count):
    """Key method example.

    Args:
        buf (str): pointer to the current WeeChat buffer.
        input_line (str): the content of the input line.
        cur (int): the position of the cursor.
        count (int): the amount of times to multiply or iterate the action.

    Notes:
        Should be called `key_X`, where X represents the key(s), and defined
        in `VI_KEYS`.
        Must perform actions on its own (using the WeeChat API).

    See Also:
        For additional examples, see `key_a()` (normal key) and
        `key_r()` (catching key).
    """
    # Key was pressed. Go to Insert mode (similar to "i").
    set_mode("INSERT")


# Operators:
# ----------

def operator_d(buf, input_line, pos1, pos2, overwrite=False):
    """Delete text from `pos1` to `pos2` from the input line.

    If `overwrite` is set to True, the character at the cursor's new position
    is removed as well (the motion is inclusive).

    See Also:
        `operator_base()`.
    """
    start = min(pos1, pos2)
    end = max(pos1, pos2)
    if overwrite:
        end += 1
    input_line = list(input_line)
    del input_line[start:end]
    input_line = "".join(input_line)
    weechat.buffer_set(buf, "input", input_line)
    set_cur(buf, input_line, pos1)

def operator_c(buf, input_line, pos1, pos2, overwrite=False):
    """Delete text from `pos1` to `pos2` from the input and enter Insert mode.

    If `overwrite` is set to True, the character at the cursor's new position
    is removed as well (the motion is inclusive.)

    See Also:
        `operator_base()`.
    """
    operator_d(buf, input_line, pos1, pos2, overwrite)
    set_mode("INSERT")

def operator_y(buf, input_line, pos1, pos2, _):
    """Yank text from `pos1` to `pos2` from the input line.

    See Also:
        `operator_base()`.
    """
    start = min(pos1, pos2)
    end = max(pos1, pos2)
    cmd = vimode_settings['copy_clipboard_cmd']
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
    proc.communicate(input=input_line[start:end].encode())


# Motions:
# --------

def motion_0(input_line, cur, count):
    """Go to the first character of the line.

    See Also;
        `motion_base()`.
    """
    return 0, False, False

def motion_w(input_line, cur, count):
    """Go `count` words forward and return position.

    See Also:
        `motion_base()`.
    """
    pos = get_pos(input_line, REGEX_MOTION_LOWERCASE_W, cur, True, count)
    if pos == -1:
        return len(input_line), False, False
    return cur + pos, False, False

def motion_W(input_line, cur, count):
    """Go `count` WORDS forward and return position.

    See Also:
        `motion_base()`.
    """
    pos = get_pos(input_line, REGEX_MOTION_UPPERCASE_W, cur, True, count)
    if pos == -1:
        return len(input_line), False, False
    return cur + pos, False, False

def motion_e(input_line, cur, count):
    """Go to the end of `count` words and return position.

    See Also:
        `motion_base()`.
    """
    for _ in range(max(1, count)):
        found = False
        pos = cur
        for pos in range(cur + 1, len(input_line) - 1):
            # Whitespace, keep going.
            if WHITESPACE.match(input_line[pos]):
                pass
            # End of sequence made from 'iskeyword' characters only,
            # or end of sequence made from non 'iskeyword' characters only.
            elif ((IS_KEYWORD.match(input_line[pos]) and
                   (not IS_KEYWORD.match(input_line[pos + 1]) or
                    WHITESPACE.match(input_line[pos + 1]))) or
                  (not IS_KEYWORD.match(input_line[pos]) and
                   (IS_KEYWORD.match(input_line[pos + 1]) or
                    WHITESPACE.match(input_line[pos + 1])))):
                found = True
                cur = pos
                break
        # We're at the character before the last and we still found nothing.
        # Go to the last character.
        if not found:
            cur = pos + 1
    return cur, True, False

def motion_E(input_line, cur, count):
    """Go to the end of `count` WORDS and return cusor position.

    See Also:
        `motion_base()`.
    """
    pos = get_pos(input_line, REGEX_MOTION_UPPERCASE_E, cur, True, count)
    if pos == -1:
        return len(input_line), False, False
    return cur + pos, True, False

def motion_b(input_line, cur, count):
    """Go `count` words backwards and return position.

    See Also:
        `motion_base()`.
    """
    # "b" is just "e" on inverted data (e.g. "olleH" instead of "Hello").
    pos_inv = motion_e(input_line[::-1], len(input_line) - cur - 1, count)[0]
    pos = len(input_line) - pos_inv - 1
    return pos, True, False

def motion_B(input_line, cur, count):
    """Go `count` WORDS backwards and return position.

    See Also:
        `motion_base()`.
    """
    new_cur = len(input_line) - cur
    pos = get_pos(input_line[::-1], REGEX_MOTION_UPPERCASE_B, new_cur,
                  count=count)
    if pos == -1:
        return 0, False, False
    pos = len(input_line) - (pos + new_cur + 1)
    return pos, True, False

def motion_ge(input_line, cur, count):
    """Go to end of `count` words backwards and return position.

    See Also:
        `motion_base()`.
    """
    # "ge is just "w" on inverted data (e.g. "olleH" instead of "Hello").
    pos_inv = motion_w(input_line[::-1], len(input_line) - cur - 1, count)[0]
    pos = len(input_line) - pos_inv - 1
    return pos, True, False

def motion_gE(input_line, cur, count):
    """Go to end of `count` WORDS backwards and return position.

    See Also:
        `motion_base()`.
    """
    new_cur = len(input_line) - cur - 1
    pos = get_pos(input_line[::-1], REGEX_MOTION_G_UPPERCASE_E, new_cur,
                  True, count)
    if pos == -1:
        return 0, False, False
    pos = len(input_line) - (pos + new_cur + 1)
    return pos, True, False

def motion_h(input_line, cur, count):
    """Go `count` characters to the left and return position.

    See Also:
        `motion_base()`.
    """
    return max(0, cur - max(count, 1)), False, False

def motion_l(input_line, cur, count):
    """Go `count` characters to the right and return position.

    See Also:
        `motion_base()`.
    """
    return cur + max(count, 1), False, False

def motion_carret(input_line, cur, count):
    """Go to first non-blank character of line and return position.

    See Also:
        `motion_base()`.
    """
    pos = get_pos(input_line, REGEX_MOTION_CARRET, 0)
    return pos, False, False

def motion_dollar(input_line, cur, count):
    """Go to end of line and return position.

    See Also:
        `motion_base()`.
    """
    pos = len(input_line)
    return pos, False, False

def motion_f(input_line, cur, count):
    """Go to `count`'th occurence of character and return position.

    See Also:
        `motion_base()`.
    """
    return start_catching_keys(1, "cb_motion_f", input_line, cur, count)

def cb_motion_f(update_last=True):
    """Callback for `motion_f()`.

    Args:
        update_last (bool, optional): should `last_search_motion` be updated?
            Set to False when calling from `key_semicolon()` or `key_comma()`
            so that the last search motion isn't overwritten.
            Defaults to True.

    See Also:
        `start_catching_keys()`.
    """
    global last_search_motion
    pattern = catching_keys_data['keys']
    pos = get_pos(catching_keys_data['input_line'], re.escape(pattern),
                  catching_keys_data['cur'], True,
                  catching_keys_data['count'])
    catching_keys_data['new_cur'] = max(0, pos) + catching_keys_data['cur']
    if update_last:
        last_search_motion = {'motion': "f", 'data': pattern}
    cb_key_combo_default(None, None, "")

def motion_F(input_line, cur, count):
    """Go to `count`'th occurence of char to the right and return position.

    See Also:
        `motion_base()`.
    """
    return start_catching_keys(1, "cb_motion_F", input_line, cur, count)

def cb_motion_F(update_last=True):
    """Callback for `motion_F()`.

    Args:
        update_last (bool, optional): should `last_search_motion` be updated?
            Set to False when calling from `key_semicolon()` or `key_comma()`
            so that the last search motion isn't overwritten.
            Defaults to True.

    See Also:
        `start_catching_keys()`.
    """
    global last_search_motion
    pattern = catching_keys_data['keys']
    cur = len(catching_keys_data['input_line']) - catching_keys_data['cur']
    pos = get_pos(catching_keys_data['input_line'][::-1],
                  re.escape(pattern),
                  cur,
                  False,
                  catching_keys_data['count'])
    catching_keys_data['new_cur'] = catching_keys_data['cur'] - max(0, pos + 1)
    if update_last:
        last_search_motion = {'motion': "F", 'data': pattern}
    cb_key_combo_default(None, None, "")

def motion_t(input_line, cur, count):
    """Go to `count`'th occurence of char and return position.

    The position returned is the position of the character to the left of char.

    See Also:
        `motion_base()`.
    """
    return start_catching_keys(1, "cb_motion_t", input_line, cur, count)

def cb_motion_t(update_last=True):
    """Callback for `motion_t()`.

    Args:
        update_last (bool, optional): should `last_search_motion` be updated?
            Set to False when calling from `key_semicolon()` or `key_comma()`
            so that the last search motion isn't overwritten.
            Defaults to True.

    See Also:
        `start_catching_keys()`.
    """
    global last_search_motion
    pattern = catching_keys_data['keys']
    pos = get_pos(catching_keys_data['input_line'], re.escape(pattern),
                  catching_keys_data['cur'] + 1,
                  True, catching_keys_data['count'])
    pos += 1
    if pos > 0:
        catching_keys_data['new_cur'] = pos + catching_keys_data['cur'] - 1
    else:
        catching_keys_data['new_cur'] = catching_keys_data['cur']
    if update_last:
        last_search_motion = {'motion': "t", 'data': pattern}
    cb_key_combo_default(None, None, "")

def motion_T(input_line, cur, count):
    """Go to `count`'th occurence of char to the left and return position.

    The position returned is the position of the character to the right of
    char.

    See Also:
        `motion_base()`.
    """
    return start_catching_keys(1, "cb_motion_T", input_line, cur, count)

def cb_motion_T(update_last=True):
    """Callback for `motion_T()`.

    Args:
        update_last (bool, optional): should `last_search_motion` be updated?
            Set to False when calling from `key_semicolon()` or `key_comma()`
            so that the last search motion isn't overwritten.
            Defaults to True.

    See Also:
        `start_catching_keys()`.
    """
    global last_search_motion
    pattern = catching_keys_data['keys']
    pos = get_pos(catching_keys_data['input_line'][::-1], re.escape(pattern),
                  (len(catching_keys_data['input_line']) -
                   (catching_keys_data['cur'] + 1)) + 1,
                  True, catching_keys_data['count'])
    pos += 1
    if pos > 0:
        catching_keys_data['new_cur'] = catching_keys_data['cur'] - pos + 1
    else:
        catching_keys_data['new_cur'] = catching_keys_data['cur']
    if update_last:
        last_search_motion = {'motion': "T", 'data': pattern}
    cb_key_combo_default(None, None, "")


# Keys:
# -----

def key_cc(buf, input_line, cur, count):
    """Delete line and start Insert mode.

    See Also:
        `key_base()`.
    """
    weechat.command("", "/input delete_line")
    set_mode("INSERT")

def key_C(buf, input_line, cur, count):
    """Delete from cursor to end of line and start Insert mode.

    See Also:
        `key_base()`.
    """
    weechat.command("", "/input delete_end_of_line")
    set_mode("INSERT")

def key_yy(buf, input_line, cur, count):
    """Yank line.

    See Also:
        `key_base()`.
    """
    cmd = vimode_settings['copy_clipboard_cmd']
    proc = subprocess.Popen(cmd, shell=True, stdin=subprocess.PIPE)
    proc.communicate(input=input_line.encode())

def key_p(buf, input_line, cur, count):
    """Paste text.

    See Also:
        `key_base()`.
    """
    cmd = vimode_settings['paste_clipboard_cmd']
    weechat.hook_process(cmd, 10 * 1000, "cb_key_p", weechat.current_buffer())

def cb_key_p(data, command, return_code, output, err):
    """Callback for fetching clipboard text and pasting it."""
    buf = ""
    this_buffer = data
    if output != "":
        buf += output.strip()
    if return_code == 0:
        my_input = weechat.buffer_get_string(this_buffer, "input")
        pos = weechat.buffer_get_integer(this_buffer, "input_pos")
        my_input = my_input[:pos] + buf + my_input[pos:]
        pos += len(buf)
        weechat.buffer_set(this_buffer, "input", my_input)
        weechat.buffer_set(this_buffer, "input_pos", str(pos))
    return weechat.WEECHAT_RC_OK

def key_i(buf, input_line, cur, count):
    """Start Insert mode.

    See Also:
        `key_base()`.
    """
    set_mode("INSERT")

def key_a(buf, input_line, cur, count):
    """Move cursor one character to the right and start Insert mode.

    See Also:
        `key_base()`.
    """
    set_cur(buf, input_line, cur + 1, False)
    set_mode("INSERT")

def key_A(buf, input_line, cur, count):
    """Move cursor to end of line and start Insert mode.

    See Also:
        `key_base()`.
    """
    set_cur(buf, input_line, len(input_line), False)
    set_mode("INSERT")

def key_I(buf, input_line, cur, count):
    """Move cursor to first non-blank character and start Insert mode.

    See Also:
        `key_base()`.
    """
    pos, _, _ = motion_carret(input_line, cur, 0)
    set_cur(buf, input_line, pos)
    set_mode("INSERT")

def key_G(buf, input_line, cur, count):
    """Scroll to specified line or bottom of buffer.

    See Also:
        `key_base()`.
    """
    if count > 0:
        # This is necessary to prevent weird scroll jumps.
        weechat.command("", "/window scroll_top")
        weechat.command("", "/window scroll %s" % (count - 1))
    else:
        weechat.command("", "/window scroll_bottom")

def key_r(buf, input_line, cur, count):
    """Replace `count` characters under the cursor.

    See Also:
        `key_base()`.
    """
    start_catching_keys(1, "cb_key_r", input_line, cur, count, buf)

def cb_key_r():
    """Callback for `key_r()`.

    See Also:
        `start_catching_keys()`.
    """
    global catching_keys_data
    input_line = list(catching_keys_data['input_line'])
    count = max(catching_keys_data['count'], 1)
    cur = catching_keys_data['cur']
    if cur + count <= len(input_line):
        for _ in range(count):
            input_line[cur] = catching_keys_data['keys']
            cur += 1
        input_line = "".join(input_line)
        weechat.buffer_set(catching_keys_data['buf'], "input", input_line)
        set_cur(catching_keys_data['buf'], input_line, cur - 1)
    catching_keys_data = {'amount': 0}

def key_R(buf, input_line, cur, count):
    """Start Replace mode.

    See Also:
        `key_base()`.
    """
    set_mode("REPLACE")

def key_tilda(buf, input_line, cur, count):
    """Switch the case of `count` characters under the cursor.

    See Also:
        `key_base()`.
    """
    input_line = list(input_line)
    count = max(1, count)
    while count and cur < len(input_line):
        input_line[cur] = input_line[cur].swapcase()
        count -= 1
        cur += 1
    input_line = "".join(input_line)
    weechat.buffer_set(buf, "input", input_line)
    set_cur(buf, input_line, cur)

def key_alt_j(buf, input_line, cur, count):
    """Go to WeeChat buffer.

    Called to preserve WeeChat's alt-j buffer switching.

    This is only called when alt-j<num> is pressed after pressing Esc, because
    \x01\x01j is received in key_combo_default which becomes \x01j after
    removing the detected Esc key.
    If Esc isn't the last pressed key, \x01j<num> is directly received in
    key_combo_default.
    """
    start_catching_keys(2, "cb_key_alt_j", input_line, cur, count)

def cb_key_alt_j():
    """Callback for `key_alt_j()`.

    See Also:
        `start_catching_keys()`.
    """
    global catching_keys_data
    weechat.command("", "/buffer " + catching_keys_data['keys'])
    catching_keys_data = {'amount': 0}

def key_semicolon(buf, input_line, cur, count, swap=False):
    """Repeat last f, t, F, T `count` times.

    Args:
        swap (bool, optional): if True, the last motion will be repeated in the
            opposite direction (e.g. "f" instead of "F"). Defaults to False.

    See Also:
        `key_base()`.
    """
    global catching_keys_data, vi_buffer
    catching_keys_data = ({'amount': 0,
                           'input_line': input_line,
                           'cur': cur,
                           'keys': last_search_motion['data'],
                           'count': count,
                           'new_cur': 0,
                           'buf': buf})
    # Swap the motion's case if called from key_comma.
    if swap:
        motion = last_search_motion['motion'].swapcase()
    else:
        motion = last_search_motion['motion']
    func = "cb_motion_%s" % motion
    vi_buffer = motion
    globals()[func](False)

def key_comma(buf, input_line, cur, count):
    """Repeat last f, t, F, T in opposite direction `count` times.

    See Also:
        `key_base()`.
    """
    key_semicolon(buf, input_line, cur, count, True)

def key_u(buf, input_line, cur, count):
    """Undo change `count` times.

    See Also:
        `key_base()`.
    """
    buf = weechat.current_buffer()
    if buf not in undo_history:
        return
    for _ in range(max(count, 1)):
        if undo_history_index[buf] > -len(undo_history[buf]):
            undo_history_index[buf] -= 1
            input_line = undo_history[buf][undo_history_index[buf]]
            weechat.buffer_set(buf, "input", input_line)
        else:
            break

def key_ctrl_r(buf, input_line, cur, count):
    """Redo change `count` times.

    See Also:
        `key_base()`.
    """
    if buf not in undo_history:
        return
    for _ in range(max(count, 1)):
        if undo_history_index[buf] < -1:
            undo_history_index[buf] += 1
            input_line = undo_history[buf][undo_history_index[buf]]
            weechat.buffer_set(buf, "input", input_line)
        else:
            break


# Vi key bindings.
# ================

# String values will be executed as normal WeeChat commands.
# For functions, see `key_base()` for reference.
VI_KEYS = {'j': "/window scroll_down",
           'k': "/window scroll_up",
           'G': key_G,
           'gg': "/window scroll_top",
           'x': "/input delete_next_char",
           'X': "/input delete_previous_char",
           'dd': "/input delete_line",
           'D': "/input delete_end_of_line",
           'cc': key_cc,
           'C': key_C,
           'i': key_i,
           'a': key_a,
           'A': key_A,
           'I': key_I,
           'yy': key_yy,
           'p': key_p,
           'gt': "/buffer -1",
           'K': "/buffer -1",
           'gT': "/buffer +1",
           'J': "/buffer +1",
           'r': key_r,
           'R': key_R,
           '~': key_tilda,
           'nt': "/bar scroll nicklist * -100%",
           'nT': "/bar scroll nicklist * +100%",
           '\x01[[A': "/input history_previous",
           '\x01[[B': "/input history_next",
           '\x01[[C': "/input move_next_char",
           '\x01[[D': "/input move_previous_char",
           '\x01[[H': "/input move_beginning_of_line",
           '\x01[[F': "/input move_end_of_line",
           '\x01[[5~': "/window page_up",
           '\x01[[6~': "/window page_down",
           '\x01[[3~': "/input delete_next_char",
           '\x01[[2~': key_i,
           '\x01M': "/input return",
           '\x01?': "/input move_previous_char",
           ' ': "/input move_next_char",
           '\x01[j': key_alt_j,
           '\x01[1': "/buffer *1",
           '\x01[2': "/buffer *2",
           '\x01[3': "/buffer *3",
           '\x01[4': "/buffer *4",
           '\x01[5': "/buffer *5",
           '\x01[6': "/buffer *6",
           '\x01[7': "/buffer *7",
           '\x01[8': "/buffer *8",
           '\x01[9': "/buffer *9",
           '\x01[0': "/buffer *10",
           '\x01^': "/input jump_last_buffer_displayed",
           '\x01D': "/window page_down",
           '\x01U': "/window page_up",
           '\x01Wh': "/window left",
           '\x01Wj': "/window down",
           '\x01Wk': "/window up",
           '\x01Wl': "/window right",
           '\x01W=': "/window balance",
           '\x01Wx': "/window swap",
           '\x01Ws': "/window splith",
           '\x01Wv': "/window splitv",
           '\x01Wq': "/window merge",
           ';': key_semicolon,
           ',': key_comma,
           'u': key_u,
           '\x01R': key_ctrl_r}

# Add alt-j<number> bindings.
for i in range(10, 99):
    VI_KEYS['\x01[j%s' % i] = "/buffer %s" % i


# Key handling.
# =============

def cb_key_pressed(data, signal, signal_data):
    """Detect potential Esc presses.

    Alt and Esc are detected as the same key in most terminals. The difference
    is that Alt signal is sent just before the other pressed key's signal.
    We therefore use a timeout (50ms) to detect whether Alt or Esc was pressed.
    """
    global last_signal_time
    last_signal_time = time.time()
    if signal_data == "\x01[":
        # In 50ms, check if any other keys were pressed. If not, it's Esc!
        weechat.hook_timer(50, 0, 1, "cb_check_esc",
                           "{:f}".format(last_signal_time))
    return weechat.WEECHAT_RC_OK

def cb_check_esc(data, remaining_calls):
    """Check if the Esc key was pressed and change the mode accordingly."""
    global esc_pressed, vi_buffer, catching_keys_data
    # Not perfect, would be better to use direct comparison (==) but that only
    # works for py2 and not for py3.
    if abs(last_signal_time - float(data)) <= 0.000001:
        esc_pressed += 1
        if mode == "SEARCH":
            weechat.command("", "/input search_stop_here")
        set_mode("NORMAL")
        # Cancel any current partial commands.
        vi_buffer = ""
        catching_keys_data = {'amount': 0}
        weechat.bar_item_update("vi_buffer")
    return weechat.WEECHAT_RC_OK

def cb_key_combo_default(data, signal, signal_data):
    """Eat and handle key events when in Normal mode, if needed.

    The key_combo_default signal is sent when a key combo is pressed. For
    example, alt-k will send the "\x01[k" signal.

    Esc is handled a bit differently to avoid delays, see `cb_key_pressed()`.
    """
    global esc_pressed, vi_buffer, cmd_compl_text, cmd_text_orig, \
           cmd_compl_pos, cmd_history_index

    # If Esc was pressed, strip the Esc part from the pressed keys.
    # Example: user presses Esc followed by i. This is detected as "\x01[i",
    # but we only want to handle "i".
    keys = signal_data
    if esc_pressed or esc_pressed == -2:
        if keys.startswith("\x01[" * esc_pressed):
            # Multiples of 3 seem to "cancel" themselves,
            # e.g. Esc-Esc-Esc-Alt-j-11 is detected as "\x01[\x01[\x01"
            # followed by "\x01[j11" (two different signals).
            if signal_data == "\x01[" * 3:
                esc_pressed = -1  # `cb_check_esc()` will increment it to 0.
            else:
                esc_pressed = 0
        # This can happen if a valid combination is started but interrupted
        # with Esc, such as Ctrl-W→Esc→w which would send two signals:
        # "\x01W\x01[" then "\x01W\x01[w".
        # In that case, we still need to handle the next signal ("\x01W\x01[w")
        # so we use the special value "-2".
        else:
            esc_pressed = -2
        keys = keys.split("\x01[")[-1]  # Remove the "Esc" part(s).
    # Ctrl-Space.
    elif keys == "\x01@":
        set_mode("NORMAL")
        return weechat.WEECHAT_RC_OK_EAT

    # Clear the undo history for this buffer on <Return>.
    if keys == "\x01M":
        buf = weechat.current_buffer()
        clear_undo_history(buf)

    # Detect imap_esc presses if any.
    if mode == "INSERT":
        imap_esc = vimode_settings['imap_esc']
        if not imap_esc:
            return weechat.WEECHAT_RC_OK
        if (imap_esc.startswith(vi_buffer) and
                imap_esc[len(vi_buffer):len(vi_buffer)+1] == keys):
            vi_buffer += keys
            weechat.bar_item_update("vi_buffer")
            weechat.hook_timer(int(vimode_settings['imap_esc_timeout']), 0, 1,
                               "cb_check_imap_esc", vi_buffer)
        elif (vi_buffer and imap_esc.startswith(vi_buffer) and
              imap_esc[len(vi_buffer):len(vi_buffer)+1] != keys):
            vi_buffer = ""
            weechat.bar_item_update("vi_buffer")
        # imap_esc sequence detected -- remove the sequence keys from the
        # Weechat input bar and enter Normal mode.
        if imap_esc == vi_buffer:
            buf = weechat.current_buffer()
            input_line = weechat.buffer_get_string(buf, "input")
            cur = weechat.buffer_get_integer(buf, "input_pos")
            input_line = (input_line[:cur-len(imap_esc)+1] +
                          input_line[cur:])
            weechat.buffer_set(buf, "input", input_line)
            set_cur(buf, input_line, cur-len(imap_esc)+1, False)
            set_mode("NORMAL")
            vi_buffer = ""
            weechat.bar_item_update("vi_buffer")
            return weechat.WEECHAT_RC_OK_EAT
        return weechat.WEECHAT_RC_OK

    # We're in Replace mode — allow "normal" key presses (e.g. "a") and
    # overwrite the next character with them, but let the other key presses
    # pass normally (e.g. backspace, arrow keys, etc).
    if mode == "REPLACE":
        if len(keys) == 1:
            weechat.command("", "/input delete_next_char")
        elif keys == "\x01?":
            weechat.command("", "/input move_previous_char")
            return weechat.WEECHAT_RC_OK_EAT
        return weechat.WEECHAT_RC_OK

    # We're catching keys! Only "normal" key presses interest us (e.g. "a"),
    # not complex ones (e.g. backspace).
    if len(keys) == 1 and catching_keys_data['amount']:
        catching_keys_data['keys'] += keys
        catching_keys_data['amount'] -= 1
        # Done catching keys, execute the callback.
        if catching_keys_data['amount'] == 0:
            globals()[catching_keys_data['callback']]()
            vi_buffer = ""
            weechat.bar_item_update("vi_buffer")
        return weechat.WEECHAT_RC_OK_EAT

    # We're in command-line mode.
    if mode == "COMMAND":
        buf = weechat.current_buffer()
        cmd_text = weechat.buffer_get_string(buf, "input")
        weechat.hook_timer(1, 0, 1, "cb_check_cmd_mode", "")
        # Return key.
        if keys == "\x01M":
            weechat.hook_timer(1, 0, 1, "cb_exec_cmd", cmd_text)
            if len(cmd_text) > 1 and (not cmd_history or
                                      cmd_history[-1] != cmd_text):
                cmd_history.append(cmd_text)
            cmd_history_index = 0
            set_mode("NORMAL")
            buf = weechat.current_buffer()
            input_line = input_line_backup[buf]['input_line']
            weechat.buffer_set(buf, "input", input_line)
            set_cur(buf, input_line, input_line_backup[buf]['cur'], False)
        # Up arrow.
        elif keys == "\x01[[A":
            if cmd_history_index > -len(cmd_history):
                cmd_history_index -= 1
                cmd_text = cmd_history[cmd_history_index]
            weechat.buffer_set(buf, "input", cmd_text)
            set_cur(buf, cmd_text, len(cmd_text), False)
        # Down arrow.
        elif keys == "\x01[[B":
            if cmd_history_index < -1:
                cmd_history_index += 1
                cmd_text = cmd_history[cmd_history_index]
            else:
                cmd_history_index = 0
                cmd_text = ":"
            weechat.buffer_set(buf, "input", cmd_text)
            set_cur(buf, cmd_text, len(cmd_text), False)
        # Tab key. No completion when searching ("/").
        elif keys == "\x01I" and cmd_text[0] == ":":
            if cmd_text_orig is None:
                input_ = list(cmd_text)
                del input_[0]
                cmd_text_orig = "".join(input_)
            cmd_compl_list = []
            for cmd in VI_COMMANDS.keys():
                if cmd.startswith(cmd_text_orig):
                    cmd_compl_list.append(cmd)
            if cmd_compl_list:
                curr_suggestion = cmd_compl_list[cmd_compl_pos]
                cmd_text = ":%s" % curr_suggestion
                cmd_compl_list[cmd_compl_pos] = weechat.string_eval_expression(
                    "${color:bold}%s${color:-bold}" % curr_suggestion,
                    {}, {}, {})
                cmd_compl_text = ", ".join(cmd_compl_list)
                cmd_compl_pos = (cmd_compl_pos + 1) % len(cmd_compl_list)
                weechat.buffer_set(buf, "input", cmd_text)
                set_cur(buf, cmd_text, len(cmd_text), False)
        # Input.
        else:
            cmd_compl_text = ""
            cmd_text_orig = None
            cmd_compl_pos = 0
        weechat.bar_item_update("cmd_completion")
        if keys in ["\x01M", "\x01[[A", "\x01[[B"]:
            cmd_compl_text = ""
            return weechat.WEECHAT_RC_OK_EAT
        else:
            return weechat.WEECHAT_RC_OK
    # Enter command mode.
    elif keys in [":", "/"]:
        if keys == "/":
            weechat.command("", "/input search_text_here")
            if not weechat.config_string_to_boolean(
                    vimode_settings['search_vim']):
                return weechat.WEECHAT_RC_OK
        else:
            buf = weechat.current_buffer()
            cur = weechat.buffer_get_integer(buf, "input_pos")
            input_line = weechat.buffer_get_string(buf, "input")
            input_line_backup[buf] = {'input_line': input_line, 'cur': cur}
            input_line = ":"
            weechat.buffer_set(buf, "input", input_line)
            set_cur(buf, input_line, 1, False)
        set_mode("COMMAND")
        cmd_compl_text = ""
        cmd_text_orig = None
        cmd_compl_pos = 0
        return weechat.WEECHAT_RC_OK_EAT

    # Add key to the buffer.
    vi_buffer += keys
    weechat.bar_item_update("vi_buffer")
    if not vi_buffer:
        return weechat.WEECHAT_RC_OK

    # Check if the keys have a (partial or full) match. If so, also get the
    # keys without the count. (These are the actual keys we should handle.)
    # After that, `vi_buffer` is only used for display purposes — only
    # `vi_keys` is checked for all the handling.
    # If no matches are found, the keys buffer is cleared.
    matched, vi_keys, count = get_keys_and_count(vi_buffer)
    if not matched:
        vi_buffer = ""
        return weechat.WEECHAT_RC_OK_EAT
    # Check if it's a command (user defined key mapped to a :cmd).
    if vi_keys.startswith(":"):
        weechat.hook_timer(1, 0, 1, "cb_exec_cmd", "{} {}".format(vi_keys,
                                                                  count))
        vi_buffer = ""
        return weechat.WEECHAT_RC_OK_EAT
    # It's a WeeChat command (user defined key mapped to a /cmd).
    if vi_keys.startswith("/"):
        weechat.command("", vi_keys)
        vi_buffer = ""
        return weechat.WEECHAT_RC_OK_EAT

    buf = weechat.current_buffer()
    input_line = weechat.buffer_get_string(buf, "input")
    cur = weechat.buffer_get_integer(buf, "input_pos")

    # It's a default mapping. If the corresponding value is a string, we assume
    # it's a WeeChat command. Otherwise, it's a method we'll call.
    if vi_keys in VI_KEYS:
        if vi_keys not in ['u', '\x01R']:
            add_undo_history(buf, input_line)
        if isinstance(VI_KEYS[vi_keys], str):
            for _ in range(max(count, 1)):
                # This is to avoid crashing WeeChat on script reloads/unloads,
                # because no hooks must still be running when a script is
                # reloaded or unloaded.
                if (VI_KEYS[vi_keys] == "/input return" and
                        input_line.startswith("/script ")):
                    return weechat.WEECHAT_RC_OK
                weechat.command("", VI_KEYS[vi_keys])
                current_cur = weechat.buffer_get_integer(buf, "input_pos")
                set_cur(buf, input_line, current_cur)
        else:
            VI_KEYS[vi_keys](buf, input_line, cur, count)
    # It's a motion (e.g. "w") — call `motion_X()` where X is the motion, then
    # set the cursor's position to what that function returned.
    elif vi_keys in VI_MOTIONS:
        if vi_keys in SPECIAL_CHARS:
            func = "motion_%s" % SPECIAL_CHARS[vi_keys]
        else:
            func = "motion_%s" % vi_keys
        end, _, _ = globals()[func](input_line, cur, count)
        set_cur(buf, input_line, end)
    # It's an operator + motion (e.g. "dw") — call `motion_X()` (where X is
    # the motion), then we call `operator_Y()` (where Y is the operator)
    # with the position `motion_X()` returned. `operator_Y()` should then
    # handle changing the input line.
    elif (len(vi_keys) > 1 and
          vi_keys[0] in VI_OPERATORS and
          vi_keys[1:] in VI_MOTIONS):
        add_undo_history(buf, input_line)
        if vi_keys[1:] in SPECIAL_CHARS:
            func = "motion_%s" % SPECIAL_CHARS[vi_keys[1:]]
        else:
            func = "motion_%s" % vi_keys[1:]
        pos, overwrite, catching = globals()[func](input_line, cur, count)
        # If it's a catching motion, we don't want to call the operator just
        # yet -- this code will run again when the motion is complete, at which
        # point we will.
        if not catching:
            oper = "operator_%s" % vi_keys[0]
            globals()[oper](buf, input_line, cur, pos, overwrite)
    # The combo isn't completed yet (e.g. just "d").
    else:
        return weechat.WEECHAT_RC_OK_EAT

    # We've already handled the key combo, so clear the keys buffer.
    if not catching_keys_data['amount']:
        vi_buffer = ""
        weechat.bar_item_update("vi_buffer")
    return weechat.WEECHAT_RC_OK_EAT

def cb_check_imap_esc(data, remaining_calls):
    """Clear the imap_esc sequence after some time if nothing was pressed."""
    global vi_buffer
    if vi_buffer == data:
        vi_buffer = ""
        weechat.bar_item_update("vi_buffer")
    return weechat.WEECHAT_RC_OK

def cb_key_combo_search(data, signal, signal_data):
    """Handle keys while search mode is active (if search_vim is enabled)."""
    if not weechat.config_string_to_boolean(vimode_settings['search_vim']):
        return weechat.WEECHAT_RC_OK
    if mode == "COMMAND":
        if signal_data == "\x01M":
            set_mode("SEARCH")
            return weechat.WEECHAT_RC_OK_EAT
    elif mode == "SEARCH":
        if signal_data == "\x01M":
            set_mode("NORMAL")
        else:
            if signal_data == "n":
                weechat.command("", "/input search_next")
            elif signal_data == "N":
                weechat.command("", "/input search_previous")
            # Start a new search.
            elif signal_data == "/":
                weechat.command("", "/input search_stop_here")
                set_mode("NORMAL")
                weechat.command("", "/input search_text_here")
            return weechat.WEECHAT_RC_OK_EAT
    return weechat.WEECHAT_RC_OK

# Callbacks.
# ==========

# Bar items.
# ----------

def cb_vi_buffer(data, item, window):
    """Return the content of the vi buffer (pressed keys on hold)."""
    return vi_buffer

def cb_cmd_completion(data, item, window):
    """Return the text of the command line."""
    return cmd_compl_text

def cb_mode_indicator(data, item, window):
    """Return the current mode (INSERT/NORMAL/REPLACE/...)."""
    return "{}{}{}{}{}".format(
        weechat.color(mode_colors[mode]),
        vimode_settings['mode_indicator_prefix'], mode,
        vimode_settings['mode_indicator_suffix'], weechat.color("reset"))

def cb_line_numbers(data, item, window):
    """Fill the line numbers bar item."""
    bar_height = weechat.window_get_integer(window, "win_chat_height")
    content = ""
    for i in range(1, bar_height + 1):
        content += "{}{}{}\n".format(vimode_settings['line_number_prefix'], i,
                                     vimode_settings['line_number_suffix'])
    return content

# Callbacks for the line numbers bar.
# ...................................

def cb_update_line_numbers(data, signal, signal_data):
    """Call `cb_timer_update_line_numbers()` when switching buffers.

    A timer is required because the bar item is refreshed before the new buffer
    is actually displayed, so ``win_chat_height`` would refer to the old
    buffer. Using a timer refreshes the item after the new buffer is displayed.
    """
    weechat.hook_timer(10, 0, 1, "cb_timer_update_line_numbers", "")
    return weechat.WEECHAT_RC_OK

def cb_timer_update_line_numbers(data, remaining_calls):
    """Update the line numbers bar item."""
    weechat.bar_item_update("line_numbers")
    return weechat.WEECHAT_RC_OK


# Config.
# -------

def cb_config(data, option, value):
    """Script option changed, update our copy."""
    option_name = option.split(".")[-1]
    if option_name in vimode_settings:
        vimode_settings[option_name] = value
    if option_name == 'user_mappings':
        load_user_mappings()
    if "_color" in option_name:
        load_mode_colors()
    return weechat.WEECHAT_RC_OK

def load_mode_colors():
    mode_colors.update({
        'NORMAL': "{},{}".format(
            vimode_settings['mode_indicator_normal_color'],
            vimode_settings['mode_indicator_normal_color_bg']),
        'INSERT': "{},{}".format(
            vimode_settings['mode_indicator_insert_color'],
            vimode_settings['mode_indicator_insert_color_bg']),
        'REPLACE': "{},{}".format(
            vimode_settings['mode_indicator_replace_color'],
            vimode_settings['mode_indicator_replace_color_bg']),
        'COMMAND': "{},{}".format(
            vimode_settings['mode_indicator_cmd_color'],
            vimode_settings['mode_indicator_cmd_color_bg']),
        'SEARCH': "{},{}".format(
            vimode_settings['mode_indicator_search_color'],
            vimode_settings['mode_indicator_search_color_bg'])
    })

def load_user_mappings():
    """Load user-defined mappings."""
    mappings = {}
    if vimode_settings['user_mappings']:
        mappings.update(json.loads(vimode_settings['user_mappings']))
    vimode_settings['user_mappings'] = mappings


# Command-line execution.
# -----------------------

def cb_exec_cmd(data, remaining_calls):
    """Translate and execute our custom commands to WeeChat command."""
    # Process the entered command.
    data = list(data)
    del data[0]
    data = "".join(data)
    # s/foo/bar command.
    if data.startswith("s/"):
        cmd = data
        parsed_cmd = next(csv.reader(StringIO(cmd), delimiter="/",
                                     escapechar="\\"))
        pattern = re.escape(parsed_cmd[1])
        repl = parsed_cmd[2]
        repl = re.sub(r"([^\\])&", r"\1" + pattern, repl)
        flag = None
        if len(parsed_cmd) == 4:
            flag = parsed_cmd[3]
        count = 1
        if flag == "g":
            count = 0
        buf = weechat.current_buffer()
        input_line = weechat.buffer_get_string(buf, "input")
        input_line = re.sub(pattern, repl, input_line, count)
        weechat.buffer_set(buf, "input", input_line)
    # Shell command.
    elif data.startswith("!"):
        weechat.command("", "/exec -buffer shell %s" % data[1:])
    # Commands like `:22`. This should start cursor mode (``/cursor``) and take
    # us to the relevant line.
    elif data.isdigit():
        line_number = int(data)
        hdata_window = weechat.hdata_get("window")
        window = weechat.current_window()
        x = weechat.hdata_integer(hdata_window, window, "win_chat_x")
        y = (weechat.hdata_integer(hdata_window, window, "win_chat_y") +
             (line_number - 1))
        weechat.command("", "/cursor go {},{}".format(x, y))
    # Check againt defined commands.
    elif data:
        raw_data = data
        data = data.split(" ", 1)
        cmd = data[0]
        args = ""
        if len(data) == 2:
            args = data[1]
        if cmd in VI_COMMANDS:
            if isinstance(VI_COMMANDS[cmd], str):
                weechat.command("", "%s %s" % (VI_COMMANDS[cmd], args))
            else:
                VI_COMMANDS[cmd](args)
        else:
            # Check for commands not sepearated by space (e.g. "b2")
            for i in range(1, len(raw_data)):
                tmp_cmd = raw_data[:i]
                tmp_args = raw_data[i:]
                if tmp_cmd in VI_COMMANDS and tmp_args.isdigit():
                    weechat.command("", "%s %s" % (VI_COMMANDS[tmp_cmd],
                                                   tmp_args))
                    return weechat.WEECHAT_RC_OK
            # No vi commands found, run the command as WeeChat command
            weechat.command("", "/{} {}".format(cmd, args))
    return weechat.WEECHAT_RC_OK

def cb_vimode_go_to_normal(data, buf, args):
    set_mode("NORMAL")
    return weechat.WEECHAT_RC_OK

# Script commands.
# ----------------

def cb_vimode_cmd(data, buf, args):
    """Handle script commands (``/vimode <command>``)."""
    # ``/vimode`` or ``/vimode help``
    if not args or args == "help":
        weechat.prnt("", "[vimode.py] %s" % README_URL)
    # ``/vimode bind_keys`` or ``/vimode bind_keys --list``
    elif args.startswith("bind_keys"):
        infolist = weechat.infolist_get("key", "", "default")
        weechat.infolist_reset_item_cursor(infolist)
        commands = ["/key unbind ctrl-W",
                    "/key bind ctrl-W /input delete_previous_word",
                    "/key bind ctrl-^ /input jump_last_buffer_displayed",
                    "/key bind ctrl-Wh /window left",
                    "/key bind ctrl-Wj /window down",
                    "/key bind ctrl-Wk /window up",
                    "/key bind ctrl-Wl /window right",
                    "/key bind ctrl-W= /window balance",
                    "/key bind ctrl-Wx /window swap",
                    "/key bind ctrl-Ws /window splith",
                    "/key bind ctrl-Wv /window splitv",
                    "/key bind ctrl-Wq /window merge"]
        while weechat.infolist_next(infolist):
            key = weechat.infolist_string(infolist, "key")
            if re.match(REGEX_PROBLEMATIC_KEYBINDINGS, key):
                commands.append("/key unbind %s" % key)
        if args == "bind_keys":
            weechat.prnt("", "Running commands:")
            for command in commands:
                weechat.command("", command)
            weechat.prnt("", "Done.")
        elif args == "bind_keys --list":
            weechat.prnt("", "Listing commands we'll run:")
            for command in commands:
                weechat.prnt("", "    %s" % command)
            weechat.prnt("", "Done.")
    return weechat.WEECHAT_RC_OK


# Helpers.
# ========

# Motions/keys helpers.
# ---------------------

def get_pos(data, regex, cur, ignore_cur=False, count=0):
    """Return the position of `regex` match in `data`, starting at `cur`.

    Args:
        data (str): the data to search in.
        regex (pattern): regex pattern to search for.
        cur (int): where to start the search.
        ignore_cur (bool, optional): should the first match be ignored if it's
            also the character at `cur`?
            Defaults to False.
        count (int, optional): the index of the match to return. Defaults to 0.

    Returns:
        int: position of the match. -1 if no matches are found.
    """
    # List of the *positions* of the found patterns.
    matches = [m.start() for m in re.finditer(regex, data[cur:])]
    pos = -1
    if count:
        if len(matches) > count - 1:
            if ignore_cur and matches[0] == 0:
                if len(matches) > count:
                    pos = matches[count]
            else:
                pos = matches[count - 1]
    elif matches:
        if ignore_cur and matches[0] == 0:
            if len(matches) > 1:
                pos = matches[1]
        else:
            pos = matches[0]
    return pos

def set_cur(buf, input_line, pos, cap=True):
    """Set the cursor's position.

    Args:
        buf (str): pointer to the current WeeChat buffer.
        input_line (str): the content of the input line.
        pos (int): the position to set the cursor to.
        cap (bool, optional): if True, the `pos` will shortened to the length
            of `input_line` if it's too long. Defaults to True.
    """
    if cap:
        pos = min(pos, len(input_line) - 1)
    weechat.buffer_set(buf, "input_pos", str(pos))

def start_catching_keys(amount, callback, input_line, cur, count, buf=None):
    """Start catching keys. Used for special commands (e.g. "f", "r").

    amount (int): amount of keys to catch.
    callback (str): name of method to call once all keys are caught.
    input_line (str): input line's content.
    cur (int): cursor's position.
    count (int): count, e.g. "2" for "2fs".
    buf (str, optional): pointer to the current WeeChat buffer.
        Defaults to None.

    `catching_keys_data` is a dict with the above arguments, as well as:
        keys (str): pressed keys will be added under this key.
        new_cur (int): the new cursor's position, set in the callback.

    When catching keys is active, normal pressed keys (e.g. "a" but not arrows)
    will get added to `catching_keys_data` under the key "keys", and will not
    be handled any further.
    Once all keys are caught, the method defined in the "callback" key is
    called, and can use the data in `catching_keys_data` to perform its action.
    """
    global catching_keys_data
    if "new_cur" in catching_keys_data:
        new_cur = catching_keys_data['new_cur']
        catching_keys_data = {'amount': 0}
        return new_cur, True, False
    catching_keys_data = ({'amount': amount,
                           'callback': callback,
                           'input_line': input_line,
                           'cur': cur,
                           'keys': "",
                           'count': count,
                           'new_cur': 0,
                           'buf': buf})
    return cur, False, True

def get_keys_and_count(combo):
    """Check if `combo` is a valid combo and extract keys/counts if so.

    Args:
        combo (str): pressed keys combo.

    Returns:
        matched (bool): True if the combo has a (partial or full) match, False
            otherwise.
        combo (str): `combo` with the count removed. These are the actual keys
            we should handle. User mappings are also expanded.
        count (int): count for `combo`.
    """
    # Look for a potential match (e.g. "d" might become "dw" or "dd" so we
    # accept it, but "d9" is invalid).
    matched = False
    # Digits are allowed at the beginning (counts or "0").
    count = 0
    if combo.isdigit():
        matched = True
    elif combo and combo[0].isdigit():
        count = ""
        for char in combo:
            if char.isdigit():
                count += char
            else:
                break
        combo = combo.replace(count, "", 1)
        count = int(count)
    # It's a user defined key. Expand it.
    if combo in vimode_settings['user_mappings']:
        combo = vimode_settings['user_mappings'][combo]
    # It's a WeeChat command.
    if not matched and combo.startswith("/"):
        matched = True
    # Check against defined keys.
    if not matched:
        for key in VI_KEYS:
            if key.startswith(combo):
                matched = True
                break
    # Check against defined motions.
    if not matched:
        for motion in VI_MOTIONS:
            if motion.startswith(combo):
                matched = True
                break
    # Check against defined operators + motions.
    if not matched:
        for operator in VI_OPERATORS:
            if combo.startswith(operator):
                # Check for counts before the motion (but after the operator).
                vi_keys_no_op = combo[len(operator):]
                # There's no motion yet.
                if vi_keys_no_op.isdigit():
                    matched = True
                    break
                # Get the motion count, then multiply the operator count by
                # it, similar to vim's behavior.
                elif vi_keys_no_op and vi_keys_no_op[0].isdigit():
                    motion_count = ""
                    for char in vi_keys_no_op:
                        if char.isdigit():
                            motion_count += char
                        else:
                            break
                    # Remove counts from `vi_keys_no_op`.
                    combo = combo.replace(motion_count, "", 1)
                    motion_count = int(motion_count)
                    count = max(count, 1) * motion_count
                # Check against defined motions.
                for motion in VI_MOTIONS:
                    if motion.startswith(combo[1:]):
                        matched = True
                        break
    return matched, combo, count


# Other helpers.
# --------------

def set_mode(arg):
    """Set the current mode and update the bar mode indicator."""
    global mode
    buf = weechat.current_buffer()
    input_line = weechat.buffer_get_string(buf, "input")
    if mode == "INSERT" and arg == "NORMAL":
        add_undo_history(buf, input_line)
    mode = arg
    # If we're going to Normal mode, the cursor must move one character to the
    # left.
    if mode == "NORMAL":
        cur = weechat.buffer_get_integer(buf, "input_pos")
        set_cur(buf, input_line, cur - 1, False)
    weechat.bar_item_update("mode_indicator")

def cb_check_cmd_mode(data, remaining_calls):
    """Exit command mode if user erases the leading ':' character."""
    buf = weechat.current_buffer()
    cmd_text = weechat.buffer_get_string(buf, "input")
    if not cmd_text:
        set_mode("NORMAL")
    return weechat.WEECHAT_RC_OK

def add_undo_history(buf, input_line):
    """Add an item to the per-buffer undo history."""
    if buf in undo_history:
        if not undo_history[buf] or undo_history[buf][-1] != input_line:
            undo_history[buf].append(input_line)
            undo_history_index[buf] = -1
    else:
        undo_history[buf] = ['', input_line]
        undo_history_index[buf] = -1

def clear_undo_history(buf):
    """Clear the undo history for a given buffer."""
    undo_history[buf] = ['']
    undo_history_index[buf] = -1

def print_warning(text):
    """Print warning, in red, to the current buffer."""
    weechat.prnt("", ("%s[vimode.py] %s" % (weechat.color("red"), text)))

def check_warnings():
    """Warn the user about problematic key bindings and tmux/screen."""
    user_warned = False
    # Warn the user about problematic key bindings that may conflict with
    # vimode.
    # The solution is to remove these key bindings, but that's up to the user.
    infolist = weechat.infolist_get("key", "", "default")
    problematic_keybindings = []
    while weechat.infolist_next(infolist):
        key = weechat.infolist_string(infolist, "key")
        command = weechat.infolist_string(infolist, "command")
        if re.match(REGEX_PROBLEMATIC_KEYBINDINGS, key):
            problematic_keybindings.append("%s -> %s" % (key, command))
    if problematic_keybindings:
        user_warned = True
        print_warning("Problematic keybindings detected:")
        for keybinding in problematic_keybindings:
            print_warning("    %s" % keybinding)
        print_warning("These keybindings may conflict with vimode.")
        print_warning("You can remove problematic key bindings and add"
                      " recommended ones by using /vimode bind_keys, or only"
                      " list them with /vimode bind_keys --list")
        print_warning("For help, see: %s" % FAQ_KEYBINDINGS)
    del problematic_keybindings
    # Warn tmux/screen users about possible Esc detection delays.
    if "STY" in os.environ or "TMUX" in os.environ:
        if user_warned:
            weechat.prnt("", "")
        user_warned = True
        print_warning("tmux/screen users, see: %s" % FAQ_ESC)
    if (user_warned and not
            weechat.config_string_to_boolean(vimode_settings['no_warn'])):
        if user_warned:
            weechat.prnt("", "")
        print_warning("To force disable warnings, you can set"
                      " plugins.var.python.vimode.no_warn to 'on'")


# Main script.
# ============

if __name__ == "__main__":
    weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION,
                     SCRIPT_LICENSE, SCRIPT_DESC, "", "")
    # Warn the user if he's using an unsupported WeeChat version.
    VERSION = weechat.info_get("version_number", "")
    if int(VERSION) < 0x01000000:
        print_warning("Please upgrade to WeeChat ≥ 1.0.0. Previous versions"
                      " are not supported.")
    # Set up script options.
    for option, value in list(vimode_settings.items()):
        if weechat.config_is_set_plugin(option):
            vimode_settings[option] = weechat.config_get_plugin(option)
        else:
            weechat.config_set_plugin(option, value[0])
            vimode_settings[option] = value[0]
        weechat.config_set_desc_plugin(option,
                                       "%s (default: \"%s\")" % (value[1],
                                                                 value[0]))
    load_user_mappings()
    load_mode_colors()
    # Warn the user about possible problems if necessary.
    if not weechat.config_string_to_boolean(vimode_settings['no_warn']):
        check_warnings()
    # Create bar items and setup hooks.
    weechat.bar_item_new("mode_indicator", "cb_mode_indicator", "")
    weechat.bar_item_new("cmd_completion", "cb_cmd_completion", "")
    weechat.bar_item_new("vi_buffer", "cb_vi_buffer", "")
    weechat.bar_item_new("line_numbers", "cb_line_numbers", "")
    weechat.bar_new("vi_line_numbers", "on", "0", "window", "", "left",
                    "vertical", "vertical", "0", "0", "default", "default",
                    "default", "0", "line_numbers")
    weechat.hook_config("plugins.var.python.%s.*" % SCRIPT_NAME, "cb_config",
                        "")
    weechat.hook_signal("key_pressed", "cb_key_pressed", "")
    weechat.hook_signal("key_combo_default", "cb_key_combo_default", "")
    weechat.hook_signal("key_combo_search", "cb_key_combo_search", "")
    weechat.hook_signal("buffer_switch", "cb_update_line_numbers", "")
    weechat.hook_command("vimode", SCRIPT_DESC, "[help | bind_keys [--list]]",
                         "     help: show help\n"
                         "bind_keys: unbind problematic keys, and bind"
                         " recommended keys to use in WeeChat\n"
                         "          --list: only list changes",
                         "help || bind_keys |--list",
                         "cb_vimode_cmd", "")
    weechat.hook_command("vimode_go_to_normal",
                         ("This command can be used for key bindings to go to "
                          "normal mode."),
                         "", "", "", "cb_vimode_go_to_normal", "")
    # Remove obsolete bar.
    vi_cmd_bar = weechat.bar_search("vi_cmd")
    weechat.bar_remove(vi_cmd_bar)



#
# Copyright (C) 2016 Andrew Rodgers-Schatz <me@andrew.rs>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from __future__ import print_function

try:
    import weechat
except Exception:
    print('This script must be run under WeeChat.')
    print('Get WeeChat now at: https://weechat.org/')
    import_ok = False

import time
import re

SCRIPT_NAME     = 'unhighlight'
SCRIPT_AUTHOR   = 'xiagu'
SCRIPT_VERSION  = '0.1.2'
SCRIPT_LICENSE  = 'GPL3'
SCRIPT_DESC     = 'Allows per-buffer specification of a regex that prevents highlights.'


def matches_unhighlight_strings(msg, regex):
    return weechat.string_has_highlight_regex(msg, regex)


def unhighlight_cb(data, modifier, modifier_data, message):
    """Check if the line matches the unhighlight regular expression, and if it does, clear the message and reprint it with the no_highlight tag added."""

    # Buffer names can have ; in them, but not plugins or tags (I HOPE),
    # so just split(';') would have weird edgecases with channels that have a ;
    # in the name.
    m = re.match("^(?P<plugin>\S+);(?P<full_name>\S+);(?P<tags>\S*)$", modifier_data)

    tags = m.group('tags')
    if 'no_highlight' in tags or 'notify_none' in tags:
        return message

    plugin = m.group('plugin')
    full_name = m.group('full_name')
    buffer = weechat.buffer_search(plugin, full_name)

    unhighlight_regex = weechat.buffer_get_string(buffer, 'localvar_unhighlight_regex')
    if not matches_unhighlight_strings(message, unhighlight_regex):
        return message

    # inspired by https://weechat.org/scripts/source/mass_hl_blocker.pl.html/
    # this is terrible and gross but afaik there is no way to change the
    # highlight message once it's set and no way to interact with a message's
    # tags before highlights are checked.
    weechat.prnt_date_tags(buffer, 0, "%s,no_highlight" % tags, message)
    return ''


def command_cb(data, buffer, args):
    args = args.strip().lower().split(' ')

    if args[0] == 'list':
        weechat.command('', '/set *.localvar_set_unhighlight_regex')
    else:
        weechat.command('', '/help %s' % SCRIPT_NAME)

    return weechat.WEECHAT_RC_OK


def main():
    hook = weechat.hook_modifier('weechat_print', 'unhighlight_cb', '')

    description = """
{script_name} lets you set up a regex for things to never highlight.

To use this, set the localvar 'unhighlight_regex' on a buffer. Lines in
that buffer which match will never be highlighted, even if they have
your nick or match highlight_words or highlight_regex.

You will need the script 'buffer_autoset.py' installed to make local
variables persistent; see the examples below.

Examples:
 Temporarily block highlights in the current buffer for lines matching 'banana':
   /buffer set localvar_set_unhighlight_regex banana
 Unhighlight SASL authentication messages for double logins:
   /buffer weechat
   /buffer set localvar_set_unhighlight_regex SaslServ
   /buffer_autoset add core.weechat localvar_set_unhighlight_regex SaslServ
 List buffers with autoset unhighlights:
   /{script_name} list
 Show this help:
   /{script_name}
 Display local variables for current buffer:
   /buffer localvar
""".format(script_name = SCRIPT_NAME)

    weechat.hook_command(SCRIPT_NAME, SCRIPT_DESC, 'list', description, 'list %-', 'command_cb', '')


if weechat.register(SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, '', ''):
    main()



"""
  :Author: Henning Hasemann <hhasemann [at] web [dot] de>

  30-01-2009 -- Adapted for Weechat 0.3.0 By Bonzodog
  (bonzodog01 [at] gmail [dot] com)

  :What it does:
    This plugin lets you inform all users in the current
    channel about the song which music-player-daemon (MPD)
    is currently playing.

  :Usage:
    /mpdnp - Display file mpd is playing to current channel.

  :Configuration Variables:
    ============= ==========================================
    Variable name Meaning
    ============= ==========================================
    host          The host where your mpd runs
    port          The port to connect to mpd (usually 6600)
    format        How this script should display
                  whats going on.
                  You may use the following variables here:
                  $artist, $title_or_file,
                  $length_min, $length_sec, $pct,
                  $pos_min, $pos_sec
  
  Released under GPL licence.
"""

todo = """
  - maybe support sending commands to mpd.
  - maybe make condional display
    (displaying some characters only
     if preceding or trailing variables are set)
"""

import weechat as wc
import mpdclient as mpd
import re
from os.path import basename, splitext

default_fmt = "/me 's MPD plays: $artist - $title_or_file ($length_min:$length_sec)"

wc.register("mpdnp", "Henning Hasemann", "0.5", "GPL", "np for mpd", "", "")

def subst(text, values):
  out = ""
  n = 0
  for match in re.finditer(findvar, text):
    if match is None: continue 
    else:
      l, r = match.span()
      nam = match.group(1)
      out += text[n:l+1] + values.get(nam, "") #"$" + nam)
      n = r
  return out + text[n:]

def np(data, buffer, args):
  """
    Send information about the currently
    played song to the channel.
  """
  host = wc.config_get_plugin("host")
  port = int(wc.config_get_plugin("port"))
  cont = mpd.MpdController(host=host, port=port)
  song = cont.getCurrentSong()
  pos, length, pct = cont.getSongPosition()
  
  # insert artist, title, album, track, path
  d = song.__dict__
  d.update({
      "title_or_file": song.title or splitext(basename(song.path))[0],
      "pos_sec": "%02d" % (pos / 60),
      "pos_min": str(pos / 60),
      "length_sec": "%02d" % (length % 60),
      "length_min": str(length / 60),
      "pct": "%2.0f" % pct,
  })
  
  wc.command(wc.current_buffer(), (subst(wc.config_get_plugin("format"), d)))
  return 0
  
def dbgnp(data, buffer, args):
  try:
    return np(data, buffer, args)
  except Exception, e:
    print e
  
wc.hook_command("mpdnp", "now playing", "", np.__doc__, "", "np", "")

findvar = re.compile(r'[^\\]\$([a-z_]+)(\b|[^a-z_])')

default = {
  "host": "localhost",
  "port": "6600",
  "format": default_fmt,
}

for k, v in default.items():
  if not wc.config_get_plugin(k):
    wc.config_set_plugin(k, v)





# -*- coding: utf-8 -*-

###
# Copyright (c) 2011, Johannes Nixdorf <mixi@shadowice.org>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
###

###
# ChangeLog:
#  0.3:
#   * don't excpect every metadata key to be available
#  0.2:
#   * fix an error when the player isn't running
#   * don't display the song if it playing is paused/stopped
#   * fix an error when the player is changed while the plugin is loaded
#   * make fewer dbus calls
#  0.1:
#   * first version
###

import dbus

try:
    import weechat
    weechat.register( 'mpris_np',
                      'Johannes Nixdorf <mixi@shadowice.org>',
                      '0.3',
                      'BSD',
                      'Print information on the currently played song',
                      '',
                      '' )
except ImportError:
    print 'This script must be called from inside weechat'

class NoPlayerException(Exception):
    pass

def getDBusName():
    dbus_root = bus.get_object( 'org.freedesktop.DBus',
                                '/' )
    dbus_if = dbus.Interface( dbus_root,
                              dbus_interface='org.freedesktop.DBus' )

    # TODO: err with multiple matches -> configurable?
    for name in dbus_if.ListNames():
        if name.startswith( 'org.mpris.' ):
            return name

    raise NoPlayerException

def getPlayerIf():
    player = bus.get_object( getDBusName(),
                             '/Player' )
    return dbus.Interface( player,
                           dbus_interface='org.freedesktop.MediaPlayer' )

def getRootIf():
    root = bus.get_object( getDBusName(),
                           '/' )
    return dbus.Interface( root,
                           dbus_interface='org.freedesktop.MediaPlayer' )

def getMetadata( name, player_if ):
    return player_if.GetMetadata()[name]

def getIdent( root_if ):
    return root_if.Identity()

def getPlayingState( player_if ):
    return player_if.GetStatus()[0]

def getPlayingStateString(x):
    if x == 0:
        return 'Playing'
    elif x == 1:
        return 'Paused'
    elif x == 2:
        return 'Stopped'

def print_info( data, buffer, args ):
    try:
        player_if = getPlayerIf()
        root_if = getRootIf()

        state = getPlayingState( player_if )
        if state == 0:
            artist = ''
            title = ''
            album = ''
            try:
                artist = getMetadata( 'artist', player_if )
            except KeyError:
                pass
            try:
                title = getMetadata( 'title', player_if )
            except KeyError:
                pass
            try:
                album = getMetadata( 'album', player_if )
            except KeyError:
                pass

            if title == '':
                string = 'np: no title available'
            else:
                string = 'np: '
                if artist != '':
                    string += '"' + artist + '" - '
                if title != '':
                    string += '"' + title + '"'
                if album != '':
                    string += ' on "' + album + '"'

            string += ' using "' + getIdent( root_if ) + '"'
        else:
            string = 'np: "' + getIdent( root_if ) + '": ' + getPlayingStateString( state )
    except NoPlayerException:
        string = 'np: no player running'

    #weechat.prnt( buffer, string )
    weechat.command( buffer, string )
    return weechat.WEECHAT_RC_OK

if __name__ == '__main__':
    global bus

    bus = dbus.SessionBus()
    weechat.hook_command( 'np',
                          'Print information on the currently played song',
                          '',
                          '',
                          '',
                          'print_info',
                          '' )




# -*- coding: utf-8 -*-
#
# Copyright (c) 2009 by xt <xt@bash.no>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

# If someone posts a spotify track URL in a configured channel
# this script will post back which track it is using spotify.url.fi service

# History:
# 2019-06-23, butlerx
#   version 0.10: Add support for spotify playlists
# 2017-09-30, butlerx
#   version 0.9: Add support for oauth keys being stored in secure data
# 2017-06-02, butlerx
#   version 0.8: add now required oauth support
# 2016-01-22, creadak
#   version 0.7: Updated for the new spotify API
# 2011-03-11, Sebastien Helleu <flashcode@flashtux.org>
#   version 0.6: get python 2.x binary for hook_process (fix problem when
#                python 3.x is default python version)
# 2010-01-12, xt
#   version 0.5: add option to use notice instead of message
# 2009-12-02, xt
#   version 0.4 small bugfix with some songs and popularity
# 2009-10-29, xt
#   version 0.3 use official spotify API, and add support for albums
# 2009-09-25, xt
#   version 0.2: use spotify.url.fi
# 2009-06-19, xt <xt@bash.no>
#     version 0.1: initial

import datetime
import re

import spotipy
import weechat as w
from spotipy.oauth2 import SpotifyClientCredentials

SCRIPT_NAME = "spotify"
SCRIPT_AUTHOR = "xt <xt@bash.no>"
SCRIPT_VERSION = "0.10"
SCRIPT_LICENSE = "GPL"
SCRIPT_DESC = "Look up spotify urls"

settings = {
    "buffers": "freenode.#mychan,",  # comma separated list of buffers
    "emit_notice": "off",  # on or off, use notice or msg
    "client_id": "client_id",
    "client_secret": "client_secret",
}

settings_help = {
    "buffers": "A comma separated list of buffers the script should check",
    "emit_notice": "If on, this script will use /notice, if off, it will use /msg to post info",
    "client_id": "required client id token go to https://developer.spotify.com/my-applications/#!/applications to generate your own",
    "client_secret": "required client secret token go to https://developer.spotify.com/my-applications/#!/applications to generate your own",
}


def parse_spotify_uri(uri):
    """
    parse spotify uri
    ---
    spotify:track:<id>
    spotify:artist:<id>
    spotify:album:<id>
    spotify:user:<user>:playlist:<id>
    """
    for regex in (
        re.compile(r"spotify:(?P<type>\w+):(?P<id>\w{22})"),
        re.compile(r"https?://open.spotify.com/(?P<type>\w+)/(?P<id>\w{22})"),
        re.compile(r"spotify:user:(?P<user>\w+):(?P<type>\w+):(?P<id>\w{22})"),
        re.compile(
            r"https?://open.spotify.com/user/(?P<user>\w+)/(?P<type>\w+)/(?P<id>\w{22})"
        ),
    ):
        results = regex.search(uri)
        if results is not None:
            yield results.groupdict()


def get_oauth(arg):
    """get oauth token from weechat conf or secure data"""
    token = w.config_get_plugin(arg)
    return (
        w.string_eval_expression(token, {}, {}, {})
        if token.startswith("${sec.data")
        else token
    )


def parse_track(data):
    """parse track data in to message"""
    name = data["name"]
    album = data["album"]["name"]
    artist = data["artists"][0]["name"]
    duration = str(datetime.timedelta(milliseconds=data["duration_ms"])).split(".")[0]
    popularity = data["popularity"]
    return "%s - %s / %s %s %d%%" % (artist, name, album, duration, popularity)


def parse_artist(data):
    """parse artist data in to message"""
    return "%s - %s followers" % (data["name"], data["followers"]["total"])


def parse_album(data):
    """parse album data in to message"""
    name = data["name"]
    artist = data["artists"][0]["name"]
    tracks = data["tracks"]["total"]
    released = data["release_date"].split("-")[0]
    length = 0
    for track in data["tracks"]["items"]:
        length += track["duration_ms"]
    duration = str(datetime.timedelta(milliseconds=length)).split(".")[0]
    return "%s - %s (%s) - %d tracks (%s)" % (artist, name, released, tracks, duration)


def parse_playlist(data):
    """parse playlist data in to message"""
    return "%s by %s - %s tracks - %s followers" % (
        data["name"],
        data["owner"]["display_name"],
        data["tracks"]["total"],
        data["followers"]["total"],
    )


def search_spotify(spotify, uri):
    """search spotify based on a uri"""
    for results in parse_spotify_uri(uri):
        if results["type"] == "album":
            yield parse_album(spotify.album(results["id"]))
        elif results["type"] == "track":
            yield parse_track(spotify.track(results["id"]))
        elif results["type"] == "artist":
            yield parse_artist(spotify.artist(results["id"]))
        elif "user" in results and results["type"] == "playlist":
            yield parse_playlist(spotify.user_playlist(results["user"], results["id"]))


def spotify_print_cb(data, buffer, time, tags, displayed, highlight, prefix, message):
    buffer_name = w.buffer_get_string(buffer, "name")
    server, channel = buffer_name.split(".")
    command = "notice" if w.config_get_plugin("emit_notice") == "on" else "msg"

    if buffer_name.lower() not in [
        buffer.lower() for buffer in w.config_get_plugin("buffers").split(",")
    ]:
        return w.WEECHAT_RC_OK

    spotify = spotipy.Spotify(
        client_credentials_manager=SpotifyClientCredentials(
            get_oauth("client_id"), get_oauth("client_secret")
        )
    )
    for reply in search_spotify(spotify, message):
        w.command("", "/%s -server %s %s %s" % (command, server, channel, reply))
    return w.WEECHAT_RC_OK


if __name__ == "__main__":
    if w.register(
        SCRIPT_NAME, SCRIPT_AUTHOR, SCRIPT_VERSION, SCRIPT_LICENSE, SCRIPT_DESC, "", ""
    ):
        # Set default settings
        for option, default in settings.items():
            if not w.config_is_set_plugin(option):
                w.config_set_plugin(option, default)

        # Set help text
        for option, description in settings_help.items():
            w.config_set_desc_plugin(option, description)

        w.hook_print("", "", "spotify", 1, "spotify_print_cb", "")



#
# Copyright (C) 2010 by Guido Berhoerster <guido+weechat@berhoerster.name>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os
import sys
import re
import string
from collections import Mapping
import weechat

SCRIPT_NAME = 'terminal-title'
VERSION = '1'
AUTHOR = 'Guido Berhoerster'
DESCRIPTION = 'Displays user defined information in the terminal title'
DEFAULT_SETTINGS = {
    'title': ('WeeChat %version [%buffer_count] %buffer_number: '
            '%buffer_name{%buffer_nicklist_count} [%hotlist]',
            'items displayed in the terminal title')
}
TERM_TEMPLATES = [
    ('xterm', "\033]0;%s\007"),
    ('screen', "\033_%s\033\\")
]
TERM_TEMPLATE = None


class TermTitleMapping(Mapping):
    substitutions = {
        'buffer_title':
                lambda : weechat.buffer_get_string(weechat.current_buffer(),
                "title"),
        'buffer_name':
                lambda : weechat.buffer_get_string(weechat.current_buffer(),
                "name"),
        'buffer_plugin':
                lambda : weechat.buffer_get_string(weechat.current_buffer(),
                "plugin"),
        'buffer_number':
                lambda : weechat.buffer_get_integer(weechat.current_buffer(),
                "number"),
        'buffer_nicklist_count':
                lambda : weechat.buffer_get_integer(weechat.current_buffer(),
                "nicklist_visible_count"),
        'buffer_count': lambda : buffer_count(),
        'hotlist': lambda : hotlist(),
        'version': lambda : weechat.info_get("version", "")
    }

    def __getitem__(self, key):
        return self.substitutions[key]()

    def __iter__(self):
        return self.substitutions.iterkeys()

    def __len__(self):
        return len(self.substitutions)


class TermTitleTemplate(string.Template):
    delimiter = '%'


def buffer_count():
    buffer_count = 0
    buffer = weechat.infolist_get("buffer", "", "")
    while weechat.infolist_next(buffer):
        buffer_count += 1
    weechat.infolist_free(buffer)
    return buffer_count

def hotlist():
    hotlist_items = []
    hotlist = weechat.infolist_get("hotlist", "", "")
    while weechat.infolist_next(hotlist):
        buffer_number = weechat.infolist_integer(hotlist, "buffer_number")
        buffer = weechat.infolist_pointer(hotlist, "buffer_pointer")
        short_name = weechat.buffer_get_string(buffer, "short_name")
        hotlist_items.append("%s:%s" % (buffer_number, short_name))
    weechat.infolist_free(hotlist)
    return ",".join(hotlist_items)

def set_term_title_hook(data, signal, signal_data):
    title_template_str = weechat.config_get_plugin('title')
    title_template = TermTitleTemplate(title_template_str)
    title_str = title_template.safe_substitute(TermTitleMapping())
    sys.__stdout__.write(TERM_TEMPLATE % title_str)
    sys.__stdout__.flush()

    return weechat.WEECHAT_RC_OK

def config_hook(data, option, value):
    set_term_title_hook("", "", "")

    return weechat.WEECHAT_RC_OK

if __name__ == '__main__':
    weechat.register(SCRIPT_NAME, AUTHOR, VERSION, 'GPL3', DESCRIPTION, '', '')

    for option, (value, description) in DEFAULT_SETTINGS.iteritems():
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, value)
        weechat.config_set_desc_plugin(option, '%s (default: "%s")' %
                (description, value))

    term = os.environ.get("TERM", None)
    if term:
        for term_name, term_template in TERM_TEMPLATES:
            if term.startswith(term_name):
                TERM_TEMPLATE = term_template
                for hook in ['buffer_switch', 'buffer_title_changed',
                        'hotlist_changed', 'upgrade_ended']:
                    weechat.hook_signal(hook, 'set_term_title_hook', '')
                weechat.hook_config('plugins.var.python.%s.*' % SCRIPT_NAME,
                        'config_hook', '')
                set_term_title_hook('', '', '')
                break



# -*- coding: utf-8 -*-
# Author: Leon Bogaert <leon AT tim-online DOT nl>
# Author: Stacey Sheldon <stac AT solidgoldbomb DOT org>
# This Plugin Calls the libindicate bindings via python when somebody says your
# nickname, sends you a query, etc.
# To make it work, you may need to download:
#    python-indicate
#    python-dbus
#    wmctrl
# Requires Weechat 0.3.0
# Released under GNU GPL v2
#
# 2010-09-22, Leon <leon@tim-online.nl>:
#     version 0.0.1 Intial release
# 2013-04-14, Stacey Sheldon <stac@solidgoldbomb.org>
#     version 0.0.2 Added two-way sync between indications and weechat
# 2013-05-01, Stacey Sheldon <stac@solidgoldbomb.org>
#     version 0.0.3 More graceful handling of missing dependencies
# 2014-05-10, Sébastien Helleu <flashcode@flashtux.org>
#     version 0.0.4 Change hook_print callback argument type of
#                   displayed/highlight (WeeChat >= 1.0)
#
# @TODO: decide what to do if a user clicks an indicator an then start typing:
#        * leave indicators alone
#        * remove indicators in the "neighbourhood"
#        * On click group: remove all indicators

import dbus.service
import os

SCRIPT_NAME    = "windicate"
SCRIPT_AUTHOR  = "Leon Bogaert"
SCRIPT_VERSION = "0.0.4"
SCRIPT_LICENSE = "GPL"
SCRIPT_DESC    = "fills the indicate applet"

DBUS_CONNECTION = 'org.weechat.scripts.windicate'
DBUS_OBJ_PATH   = '/org/weechat/scripts/windicate'

class DBUSService(dbus.service.Object):
    def __init__(self, messageMenu):
        bus_name = dbus.service.BusName(DBUS_CONNECTION, bus=dbus.SessionBus())
        dbus.service.Object.__init__(self, bus_name, DBUS_OBJ_PATH)
        self.messageMenu = messageMenu

    @dbus.service.method('org.weechat.scripts.windicate')
    def add_message(self, buffer, brief, sender, body):
        return self.messageMenu.add_message(buffer, brief, sender, body)

    @dbus.service.method('org.weechat.scripts.windicate')
    def del_messages(self, buffer):
        return self.messageMenu.del_messages(buffer)

class MessageMenu(object):
    def __init__(self, weechat_fifo, weechat_windowid):
        self.messages = []

        self.weechat_fifo = weechat_fifo
        self.weechat_windowid = weechat_windowid

        server = pyindicate.indicate_server_ref_default()
        server.set_type("message.im")
        server.set_desktop_file(self.desktop_file())
        server.connect("server-display", self.server_click)
        server.show()

    def desktop_file(self):
        DESKTOP_ENTRY = """\
[Desktop Entry]
Encoding=UTF-8
MultipleArgs=false
Terminal=true
Exec=weechat-curses
Icon=weechat
Type=Application
Categories=Network;IRCClient;
StartupNotify=false
Name=Weechat
GenericName=IRC Client
"""
        file = "/usr/share/applications/weechat.desktop"

        if os.path.isfile(file):
            return file

        import tempfile
        f = tempfile.NamedTemporaryFile(suffix='indicator', delete=False)
        f.write(DESKTOP_ENTRY)
        f.close()
        self.tmp_file = f

        return f.name

    def __del__(self):
        if self.tmp_file and os.file.exists(self.tmp_file):
            os.unlink(self.tmp_file)

    def raise_window(self):
        import subprocess
        try:
            retcode = subprocess.call('wmctrl -i -a %s' % self.weechat_windowid, shell=True)
            if retcode < 0:
                # print >>sys.stderr, "Child was terminated by signal", -retcode
                pass
            else:
                # print >>sys.stderr, "Child returned", retcode
                pass
        except OSError as e:
            # print >>sys.stderr, "Execution failed:", e
            pass

    def server_click(self, server, time):
        # tell weechat to select the first active buffer
        with open (self.weechat_fifo, 'a') as f:
            f.write ("irc.server.freenode */input jump_smart\n")
        # raise the weechat window to the foreground
        self.raise_window()

    def add_message(self, buffer, brief, sender, body):
        for message in self.messages:
            if message.buffer == buffer and message.sender == sender:
                return message.update_time()

        return self.messages.append(Message(self, buffer, brief, sender, body))

    def del_messages(self, buffer):
        # remove all pending indications for this buffer
        for message in self.messages:
            if message.buffer == buffer:
                self.messages.remove(message)
                message.indicator.hide()
        return True

class Message(object):
    def __init__(self, mm, buffer, brief, sender, message):
        # Setup the message
        try:
            # Ubuntu 9.10 and above
            indicator = pyindicate.Indicator()
        except:
            # Ubuntu 9.04
            indicator = pyindicate.IndicatorMessage()

        indicator.set_property("subtype", "im")
        indicator.set_property("sender", "%s (%s)" % (sender, brief))
        indicator.set_property("body", message)
        indicator.set_property_time("time", time())
        indicator.set_property('draw-attention', 'true');
        indicator.show()
        indicator.connect("user-display", self.message_clicked)

        self.indicator = indicator
        self.sender = sender
        self.message = message
        self.brief = brief
        self.messageMenu = mm
        self.buffer = buffer

    def update_time(self):
        self.indicator.set_property_time("time", time())

    def message_clicked(self, indicator, time):
        self.messageMenu.messages.remove(self)
        indicator.hide()
        # tell weechat to select the buffer that triggered this indication
        with open (self.messageMenu.weechat_fifo, 'a') as f:
            f.write ("core.weechat */buffer %s\n" % self.buffer)
        # raise the weechat window to the foreground
        self.messageMenu.raise_window()

class WindicateServer(object):
    def __init__(self, weechat_fifo, weechat_windowid):
        DBusGMainLoop(set_as_default=True)
        mm = MessageMenu(weechat_fifo, weechat_windowid)
        dbs = DBUSService(mm)

        loop = gobject.MainLoop()
        loop.run()

class Subprocess(object):
    p = None

    @classmethod
    def start(cls, fifo_filename, window_windowid):
        import inspect
        file = inspect.getfile(inspect.currentframe())

        import subprocess
        args = ["/usr/bin/python",
                file,
                fifo_filename,
                weechat_windowid]
        cls.p = subprocess.Popen(args)

    @classmethod
    def stop(cls):
        if cls.p != None:
            cls.p.terminate()

try:
    import weechat
except ImportError:
    """Ran from seperate process: start server!"""
    import gobject
    gobject.threads_init()
    import indicate as pyindicate
    import gtk
    import dbus
    from dbus.mainloop.glib import DBusGMainLoop
    from time import time
    import sys # argv

    # arguments: weechat_fifo_filename weechat_windowid
    ws = WindicateServer(sys.argv[1], sys.argv[2])    # doesn't return

##### FUNCTIONS #####

def weechat_script_end():
    Subprocess.stop()
    return weechat.WEECHAT_RC_OK

def notify_msg(data, bufferp, time, tags, display, is_hilight, prefix, msg):
    """Sends highlighted message to be printed on notification"""

    if ('notify_private' in tags and
        weechat.config_get_plugin('show_priv_msg') == "on") \
        or (int(is_hilight) and \
        weechat.config_get_plugin('show_hilights') == "on"):

        # grab the fully qualified buffer name so we can jump to it later
        buffer = weechat.buffer_get_string(bufferp, "name")

        # choose an appropriate brief name to display in the indicator applet
        if 'notify_private' in tags:
            brief = "private"
        else:
            # prefer short_name
            brief = weechat.buffer_get_string(bufferp, "short_name")
            if not brief:
                # fall back to full name
                brief = buffer

        if weechat.config_get_plugin('debug') == "on":
            print "buffer: " + buffer
            print "brief: " + brief
            print "prefix: " + prefix
            print "msg: " + msg

        # Create an object that will proxy for a particular remote object.
        bus = dbus.SessionBus()
        remote_object = bus.get_object(DBUS_CONNECTION, DBUS_OBJ_PATH)
        remote_object.add_message(buffer, brief, prefix, msg)

    return weechat.WEECHAT_RC_OK

def buffer_switched(data, signal, signal_data):
    buffer = weechat.buffer_get_string(signal_data, "name")
    if weechat.config_get_plugin('debug') == "on":
        print "data: " + data
        print "signal: " + signal
        print "message: " + signal_data
        print "buffer: " + buffer

    # Create an object that will proxy for a particular remote object.
    bus = dbus.SessionBus()
    remote_object = bus.get_object(DBUS_CONNECTION, DBUS_OBJ_PATH)
    remote_object.del_messages(buffer)

    return weechat.WEECHAT_RC_OK

##### END FUNCTIONS #####

settings = {
    'show_hilights' : ('on', 'Should hilights trigger indications' ),
    'show_priv_msg' : ('on', 'Should privmsgs trigger indications' ),
}

if weechat.register(SCRIPT_NAME,
                    SCRIPT_AUTHOR,
                    SCRIPT_VERSION,
                    SCRIPT_LICENSE,
                    SCRIPT_DESC,
                    "weechat_script_end",
                    ""):
    version = weechat.info_get('version_number', '') or 0

    # Init everything
    for option, default_desc in settings.items():
        if not weechat.config_is_set_plugin(option):
            weechat.config_set_plugin(option, default_desc[0])
        if int(version) >= 0x00030500:
            weechat.config_set_desc_plugin(option, default_desc[1])

    # Perform some sanity checks to make sure we have everything we need to run
    sanity = True
    weechat_windowid = os.environ.get('WINDOWID')
    if weechat_windowid == None:
        weechat.prnt("", "%sEnvironment variable WINDOWID not set.  This script requires an X environment to run." % weechat.prefix("error"))
        sanity = False

    fifo_filename = weechat.info_get("fifo_filename", "")
    if fifo_filename == "":
        weechat.prnt("", "%sWeechat variable fifo_filename is not set.  Is the fifo plugin enabled?" % weechat.prefix("error"))
        sanity = False

    if sanity:
        Subprocess.start(fifo_filename, weechat_windowid)

        # Hook privmsg/hilights
        weechat.hook_print("", "", "", 1, "notify_msg", "")
        weechat.hook_signal("buffer_switch", "buffer_switched", "")
